!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Mathematical and Computational Sciences Division                !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

module linked_list

!----------------------------------------------------
! This module implements linked lists.
!----------------------------------------------------

implicit none
private

!===============================================================================
!   Description of public entities

public &

! Structure for a linked list containing an integer for the value

   linked_list_integer, &

! Structure for a linked list containing an array of integers for the value

   linked_list_integer_array, &

! subroutine linked_list_create(list_head)
! type(linked_list_<type>), intent(out) :: list_head
! Initialize list_head to be the head of an empty linked list.

   linked_list_create, &

! subroutine linked_list_destroy(list_head)
! type(linked_list_<type>), intent(inout) :: list_head
! Destroy the linked list that starts at list_head, to free all memory
! associated with it.

   linked_list_destroy, &

! subroutine linked_list_copy(old_list,new_list)
! type(linked_list_<type>), intent(in) :: old_list
! type(linked_list_<type>), intent(inout) :: new_list
! Copy old_list to new_list.  new_list should be created and empty.

   linked_list_copy, &

! subroutine linked_list_add(list_head,new_value,follow_value)
! type(linked_list_<type>), intent(inout) :: list_head
! <type>, intent(in) :: new_value, follow_value
! Add a new entry to the linked list with value new_value.  Place it in the
! list immediately after the first entry with value follow_value.  If there is
! no such entry, it goes at the end of the list.  To force it at the end of
! the list, use LINKED_LIST_END_<type> for follow_value.  To force it at the
! beginning of the list, use LINKED_LIST_HEAD_INTEGER for follow_value.
! If <type> is an array, the order of the entries in follow_value need not be
! the same as the order in the value in the linked list.

   linked_list_add, &

! subroutine linked_list_remove(list_head,value)
! type(linked_list_<type>), intent(inout) :: list_head
! <type>, intent(in) :: value
! Remove the first entry in the linked list with the given value.  If no such
! entry exists, the list is unchanged.  If <type> is an array, the order of
! the entries in the array does not matter.

   linked_list_remove, &

! function linked_list_contains(list_head,value)
! type(linked_list_<type>), intent(in) :: list_head
! <type>, intent(in) :: value
! logical :: linked_list_contains
! Returns true if the linked list contains an entry with the given value.  If
! the <type> is an array, the order of the entries in the array does not matter.

   linked_list_contains, &

! function linked_list_length(list_head)
! type(linked_list_<type>), intent(in) :: list_head
! integer :: linked_list_length
! Returns the number of entries in the linked list

   linked_list_length, &

! function linked_list_value(list_entry)
! type(linked_list_<type>), intent(in) :: list_entry
! <type> :: linked_list_value
! Returns the value associated with list_entry.  If the list is empty it
! returns LINKED_LIST_END_<type>.

   linked_list_value, &

! function linked_list_next(list_entry)
! type(linked_list_<type>), intent(in) :: list_entry
! linked_list_<type>, pointer :: linked_list_next
! Returns a pointer to the entry that follows list_entry in the linked list.
! If list_entry is the tail, the result is null, i.e. not associated.
! To traverse a linked list, do something like
!   entry => linked list head
!   do while (associated(entry))
!      entry => linked_list_next(entry)
!   end do

   linked_list_next, &

! subroutine linked_list_print(list_head)
! type(linked_list_<type>), intent(in) :: list_head
! Prints the list of entries in a linked list

   linked_list_print, &

! Values returned by next_linked_list when there are no more entries, or used
! to indicate insertion at the head or tail.

   LINKED_LIST_END_INTEGER,  LINKED_LIST_END_INTEGER_ARRAY, &
   LINKED_LIST_HEAD_INTEGER, LINKED_LIST_HEAD_INTEGER_ARRAY

! end of public entities
!===============================================================================

!----------------------------------------------------
! The following types are defined:

type linked_list_integer
   integer :: value
   type(linked_list_integer), pointer :: next
end type linked_list_integer

type linked_list_integer_array
   integer, allocatable :: value(:)
   type(linked_list_integer_array), pointer :: next
end type linked_list_integer_array

!----------------------------------------------------
! The following defined constants are defined:

integer, parameter :: LINKED_LIST_END_INTEGER = huge(0), &
                      LINKED_LIST_END_INTEGER_ARRAY(1) = (/huge(0)/), &
                      LINKED_LIST_HEAD_INTEGER = -huge(0), &
                      LINKED_LIST_HEAD_INTEGER_ARRAY(1) = (/-huge(0)/)

!----------------------------------------------------
! The following generic interfaces are defined:

interface linked_list_create
   module procedure linked_list_integer_create, &
                    linked_list_integer_array_create
end interface

interface linked_list_destroy
   module procedure linked_list_integer_destroy, &
                    linked_list_integer_array_destroy
end interface

interface linked_list_copy
   module procedure linked_list_integer_copy, &
                    linked_list_integer_array_copy
end interface

interface linked_list_add
   module procedure linked_list_integer_add, &
                    linked_list_integer_array_add
end interface

interface linked_list_remove
   module procedure linked_list_integer_remove, &
                    linked_list_integer_array_remove
end interface

interface linked_list_contains
   module procedure linked_list_integer_contains, &
                    linked_list_integer_array_contains
end interface

interface linked_list_length
   module procedure linked_list_integer_length, &
                    linked_list_integer_array_length
end interface

interface linked_list_value
   module procedure linked_list_integer_value, &
                    linked_list_integer_array_value
end interface

interface linked_list_next
   module procedure linked_list_integer_next, &
                    linked_list_integer_array_next
end interface

interface linked_list_print
   module procedure linked_list_integer_print, &
                    linked_list_integer_array_print
end interface

!----------------------------------------------------
! The following interface operators are defined:

interface operator(.eqarray.)
   module procedure integer_array_equal
end interface

interface operator(.neqarray.)
   module procedure integer_array_notequal
end interface

!----------------------------------------------------
! Options to trace operations for debugging.

logical, parameter :: trace_integer = .false.
logical, parameter :: trace_integer_array = .false.

integer, parameter :: outunit = 6
!----------------------------------------------------

contains

!          --------------------------
subroutine linked_list_integer_create(list_head)
!          --------------------------

!----------------------------------------------------
! Initialize list_head to be the head of an empty linked list.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer), intent(out) :: list_head
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (trace_integer) write(outunit,*) "create integer linked list"

list_head%value = LINKED_LIST_END_INTEGER
nullify(list_head%next)

end subroutine linked_list_integer_create

!          --------------------------------
subroutine linked_list_integer_array_create(list_head)
!          --------------------------------

!----------------------------------------------------
! Initialize list_head to be the head of an empty linked list.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer_array), intent(out) :: list_head
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (trace_integer_array) write(outunit,*) "create integer array linked list"

allocate(list_head%value(1))
list_head%value = LINKED_LIST_END_INTEGER_ARRAY
nullify(list_head%next)

end subroutine linked_list_integer_array_create

!          ---------------------------
subroutine linked_list_integer_destroy(list_head)
!          ---------------------------

!----------------------------------------------------
! Destroy the linked list that starts at list_head, to free all memory
! associated with it.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer), intent(inout) :: list_head
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (trace_integer) write(outunit,*) "destroy integer linked list"

call integer_destroy_recursive(list_head)
list_head%value = LINKED_LIST_END_INTEGER

end subroutine linked_list_integer_destroy

!          ---------------------------------
subroutine linked_list_integer_array_destroy(list_head)
!          ---------------------------------

!----------------------------------------------------
! Destroy the linked list that starts at list_head, to free all memory
! associated with it.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer_array), intent(inout) :: list_head
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (trace_integer_array) write(outunit,*) "destroy integer array linked list"

call integer_array_destroy_recursive(list_head)
allocate(list_head%value(1))
list_head%value = LINKED_LIST_END_INTEGER_ARRAY

end subroutine linked_list_integer_array_destroy

!          ------------------------
subroutine linked_list_integer_copy(old_list,new_list)
!          ------------------------

!----------------------------------------------------
! Copy old_list to new_list.  new_list should be created and empty.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer), intent(in), target :: old_list
type(linked_list_integer), intent(inout), target :: new_list
!----------------------------------------------------
! Local variables:

type(linked_list_integer), pointer :: old_list_entry, new_list_entry
!----------------------------------------------------
! Begin executable code

if (trace_integer) then
   write(outunit,*) "copy integer linked list containing:"
   call linked_list_print(old_list)
endif

old_list_entry => old_list
new_list_entry => new_list
do
   new_list_entry%value = old_list_entry%value
   old_list_entry => old_list_entry%next
   if (.not. associated(old_list_entry)) exit
   allocate(new_list_entry%next)
   new_list_entry => new_list_entry%next
end do
nullify(new_list_entry%next)

end subroutine linked_list_integer_copy

!          ------------------------------
subroutine linked_list_integer_array_copy(old_list,new_list)
!          ------------------------------

!----------------------------------------------------
! Copy old_list to new_list.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer_array), intent(in), target :: old_list
type(linked_list_integer_array), intent(inout), target :: new_list
!----------------------------------------------------
! Local variables:

type(linked_list_integer_array), pointer :: old_list_entry, new_list_entry
!----------------------------------------------------
! Begin executable code

if (trace_integer_array) then
   write(outunit,*) "copy integer array linked list containing:"
   call linked_list_print(old_list)
endif

old_list_entry => old_list
new_list_entry => new_list
do
   allocate(new_list_entry%value(size(old_list_entry%value)))
   new_list_entry%value = old_list_entry%value
   old_list_entry => old_list_entry%next
   if (.not. associated(old_list_entry)) exit
   allocate(new_list_entry%next)
   new_list_entry => new_list_entry%next
end do
nullify(new_list_entry%next)

end subroutine linked_list_integer_array_copy

!          -----------------------
subroutine linked_list_integer_add(list_head,new_value,follow_value)
!          -----------------------

!----------------------------------------------------
! Add a new entry to the linked list with value new_value.  Place it in the
! list immediately after the first entry with value follow_value.  If there is
! no such entry, it goes at the end of the list.  To force it at the end of
! the list, use LINKED_LIST_END_INTEGER for follow_value.  To force it at the
! beginning of the list, use LINKED_LIST_HEAD_INTEGER for follow_value.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer), intent(inout), target :: list_head
integer, intent(in) :: new_value, follow_value
!----------------------------------------------------
! Local variables:

type(linked_list_integer), pointer :: entry, hold_next
!----------------------------------------------------
! Begin executable code

if (trace_integer) then
   write(outunit,*) "add ",new_value," after ",follow_value," in integer linked list containing:"
   call linked_list_print(list_head)
endif

! If the list is empty, just set the head

if (list_head%value == LINKED_LIST_END_INTEGER) then
   list_head%value = new_value

else

! If adding it as a head and the list is not empty, insert a new entry after
! the current head, and swap their values later.

   if (follow_value == LINKED_LIST_HEAD_INTEGER) then
      entry => list_head

! If not adding it as a head, search the list for follow_value; if it is not
! there we end up pointing to the tail

   else
      entry => list_head
      do while (associated(entry%next))
         if (entry%value == follow_value) exit
         entry => entry%next
      end do
   endif

! Create the new entry

! in the interior of the list

   if (associated(entry%next)) then
      hold_next => entry%next
      nullify(entry%next)
      allocate(entry%next)
      entry%next%value = new_value
      entry%next%next => hold_next

! at the tail of the list

   else
      allocate(entry%next)
      entry%next%value = new_value
      nullify(entry%next%next)
   endif

! If adding it as a head, we just inserted it after the head, so swap the old
! head and new entry

   if (follow_value == LINKED_LIST_HEAD_INTEGER) then
      list_head%next%value = list_head%value
      list_head%value = new_value
   endif

endif

end subroutine linked_list_integer_add


!          -----------------------------
subroutine linked_list_integer_array_add(list_head,new_value,follow_value)
!          -----------------------------

!----------------------------------------------------
! Add a new entry to the linked list with value new_value.  Place it in the
! list immediately after the first entry with value follow_value.  If there is
! no such entry, it goes at the end of the list.  To force it at the end of
! the list, use LINKED_LIST_END_INTEGER_ARRAY for follow_value.  To force it at
! the beginning of the list, use LINKED_LIST_HEAD_INTEGER_ARRAY for
! follow_value.  The order of the entries in follow_value need not be the same
! as the order in the value in the linked list.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer_array), intent(inout), target :: list_head
integer, intent(in) :: new_value(:), follow_value(:)
!----------------------------------------------------
! Local variables:

type(linked_list_integer_array), pointer :: entry, hold_next
!----------------------------------------------------
! Begin executable code

if (trace_integer_array) then
   write(outunit,*) "add ",new_value," after ",follow_value," in integer array linked list containing:"
   call linked_list_print(list_head)
endif

! If the list is empty, just set the head

if (list_head%value .eqarray. LINKED_LIST_END_INTEGER_ARRAY) then
   deallocate(list_head%value)
   allocate(list_head%value(size(new_value)))
   list_head%value = new_value

else

! If adding it as a head and the list is not empty, insert a new entry after
! the current head, and swap their values later.

   if (follow_value .eqarray. LINKED_LIST_HEAD_INTEGER_ARRAY) then
      entry => list_head

! If not adding it as a head, search the list for follow_value; if it is not
! there we end up pointing to the tail

   else
      entry => list_head
      do while (associated(entry%next))
         if (entry%value .eqarray. follow_value) exit
         entry => entry%next
      end do
   endif

! Create the new entry

! in the interior of the list

   if (associated(entry%next)) then
      hold_next => entry%next
      nullify(entry%next)
      allocate(entry%next)
      allocate(entry%next%value(size(new_value)))
      entry%next%value = new_value
      entry%next%next => hold_next

! at the tail of the list

   else
      allocate(entry%next)
      allocate(entry%next%value(size(new_value)))
      entry%next%value = new_value
      nullify(entry%next%next)
   endif

! If adding it as a head, we just inserted it after the head, so swap the old
! head and new entry

   if (follow_value .eqarray. LINKED_LIST_HEAD_INTEGER_ARRAY) then
      if (size(list_head%next%value) /= size(list_head%value)) then
         deallocate(list_head%next%value)
         allocate(list_head%next%value(size(list_head%value)))
      endif
      list_head%next%value = list_head%value
      if (size(list_head%value) /= size(new_value)) then
         deallocate(list_head%value)
         allocate(list_head%value(size(new_value)))
      endif
      list_head%value = new_value
   endif

endif

end subroutine linked_list_integer_array_add

!          --------------------------
subroutine linked_list_integer_remove(list_head,value)
!          --------------------------

!----------------------------------------------------
! Remove the first entry in the linked list with the given value.  If no such
! entry exists, the list is unchanged.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer), intent(inout), target :: list_head
integer, intent(in) :: value
!----------------------------------------------------
! Local variables:

type(linked_list_integer), pointer :: entry, hold_next, previous
!----------------------------------------------------
! Begin executable code

if (trace_integer) then
   write(outunit,*) "remove ",value," from integer linked list containing:"
   call linked_list_print(list_head)
endif

! If the list is empty, nothing to do

if (list_head%value /= LINKED_LIST_END_INTEGER) then

! If the list has one entry, set the list empty if the value matches, or
! leave it alone if it does not.

   if (.not. associated(list_head%next)) then

      if (list_head%value == value) then
         list_head%value = LINKED_LIST_END_INTEGER
      endif

   else

! If the list has more than one entry and the head matches the value, move the
! second entry into the head

      if (list_head%value == value) then
         list_head%value = list_head%next%value
         if (associated(list_head%next%next)) then
            hold_next => list_head%next%next
            deallocate(list_head%next)
            list_head%next => hold_next
         else
            deallocate(list_head%next)
            nullify(list_head%next)
         endif

! otherwise search the list

      else

         previous => list_head
         entry => list_head%next
         do while (associated(entry))
            if (entry%value == value) then
               if (associated(entry%next)) then
                  deallocate(previous%next)
                  previous%next => entry%next
               else
                  deallocate(previous%next)
                  nullify(previous%next)
               endif
               exit
            endif
            previous => entry
            entry => entry%next
         end do

      endif
   endif
endif

end subroutine linked_list_integer_remove

!          --------------------------------
subroutine linked_list_integer_array_remove(list_head,value)
!          --------------------------------

!----------------------------------------------------
! Remove the first entry in the linked list with the given value.  If no such
! entry exists, the list is unchanged.  The order of the entries in the array
! does not matter.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer_array), intent(inout), target :: list_head
integer, intent(in) :: value(:)
!----------------------------------------------------
! Local variables:

type(linked_list_integer_array), pointer :: entry, hold_next, previous
!----------------------------------------------------
! Begin executable code

if (trace_integer_array) then
   write(outunit,*) "remove ",value," from integer array linked list containing:"
   call linked_list_print(list_head)
endif

! If the list is empty, nothing to do

if (list_head%value .neqarray. LINKED_LIST_END_INTEGER_ARRAY) then

! If the list has one entry, set the list empty if the value matches, or
! leave it alone if it does not.

   if (.not. associated(list_head%next)) then

      if (list_head%value .eqarray. value) then
         if (size(list_head%value) /= size(LINKED_LIST_END_INTEGER_ARRAY)) then
            deallocate(list_head%value)
            allocate(list_head%value(size(LINKED_LIST_END_INTEGER_ARRAY)))
         endif
         list_head%value = LINKED_LIST_END_INTEGER_ARRAY
      endif

   else

! If the list has more than one entry and the head matches the value, move the
! second entry into the head

      if (list_head%value .eqarray. value) then
         if (size(list_head%value) /= size(list_head%next%value)) then
            deallocate(list_head%value)
            allocate(list_head%value(size(list_head%next%value)))
         endif
         list_head%value = list_head%next%value
         deallocate(list_head%next%value)
         if (associated(list_head%next%next)) then
            hold_next => list_head%next%next
            deallocate(list_head%next)
            list_head%next => hold_next
         else
            deallocate(list_head%next)
            nullify(list_head%next)
         endif

      else

         previous => list_head
         entry => list_head%next
         do while (associated(entry))
            if (entry%value .eqarray. value) then
               if (associated(entry%next)) then
                  deallocate(entry%value)
                  deallocate(previous%next)
                  previous%next => entry%next
               else
                  deallocate(entry%value)
                  deallocate(previous%next)
                  nullify(previous%next)
               endif
               exit
            endif
            previous => entry
            entry => entry%next
         end do

      endif
   endif
endif

end subroutine linked_list_integer_array_remove

!        ----------------------------
function linked_list_integer_contains(list_head,value)
!        ----------------------------

!----------------------------------------------------
! Returns true if the linked list contains an entry with the given value.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer), intent(in), target :: list_head
integer, intent(in) :: value
logical :: linked_list_integer_contains
!----------------------------------------------------
! Local variables:

type(linked_list_integer), pointer :: entry
!----------------------------------------------------
! Begin executable code

if (trace_integer) then
   write(outunit,*) "see if ",value," is in integer linked list containing:"
   call linked_list_print(list_head)
endif

if (list_head%value == LINKED_LIST_END_INTEGER) then
   linked_list_integer_contains = .false.
else
   entry => list_head
   linked_list_integer_contains = .false.
   do while (associated(entry))
      if (entry%value == value) then
         linked_list_integer_contains = .true.
         exit
      endif
      entry => entry%next
   end do
endif

end function linked_list_integer_contains

!        ----------------------------------
function linked_list_integer_array_contains(list_head,value)
!        ----------------------------------

!----------------------------------------------------
! Returns true if the linked list contains an entry with the given value.
! The order of the entries in the array does not matter.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer_array), intent(in), target :: list_head
integer, intent(in) :: value(:)
logical :: linked_list_integer_array_contains
!----------------------------------------------------
! Local variables:

type(linked_list_integer_array), pointer :: entry
!----------------------------------------------------
! Begin executable code

if (trace_integer_array) then
   write(outunit,*) "see if ",value," is in integer array linked list containing:"
   call linked_list_print(list_head)
endif

if (list_head%value .eqarray. LINKED_LIST_END_INTEGER_ARRAY) then
   linked_list_integer_array_contains = .false.
else
   entry => list_head
   linked_list_integer_array_contains = .false.
   do while (associated(entry))
      if (entry%value .eqarray. value) then
         linked_list_integer_array_contains = .true.
         exit
      endif
      entry => entry%next
   end do
endif

end function linked_list_integer_array_contains

!        --------------------------
function linked_list_integer_length(list_head)
!        --------------------------

!----------------------------------------------------
! Returns the number of entries in the linked list
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer), intent(in), target :: list_head
integer :: linked_list_integer_length
!----------------------------------------------------
! Local variables:

type(linked_list_integer), pointer :: entry
!----------------------------------------------------
! Begin executable code

if (list_head%value == LINKED_LIST_END_INTEGER) then
   linked_list_integer_length = 0
else
   entry => list_head
   linked_list_integer_length = 1
   do while (associated(entry%next))
      linked_list_integer_length = linked_list_integer_length + 1
      entry => entry%next
   end do
endif

if (trace_integer) then
   write(outunit,*) linked_list_integer_length," is the length of integer linked list containing:"
   call linked_list_print(list_head)
endif

end function linked_list_integer_length

!        --------------------------------
function linked_list_integer_array_length(list_head)
!        --------------------------------

!----------------------------------------------------
! Returns the number of entries in the linked list
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer_array), intent(in), target :: list_head
integer :: linked_list_integer_array_length
!----------------------------------------------------
! Local variables:

type(linked_list_integer_array), pointer :: entry
!----------------------------------------------------
! Begin executable code

if (list_head%value .eqarray. LINKED_LIST_END_INTEGER_ARRAY) then
   linked_list_integer_array_length = 0
else
   entry => list_head
   linked_list_integer_array_length = 1
   do while (associated(entry%next))
      linked_list_integer_array_length = linked_list_integer_array_length + 1
      entry => entry%next
   end do
endif

if (trace_integer_array) then
   write(outunit,*) linked_list_integer_array_length," is the length of integer array linked list containing:"
   call linked_list_print(list_head)
endif

end function linked_list_integer_array_length

!        -------------------------
function linked_list_integer_value(list_entry)
!        -------------------------

!----------------------------------------------------
! Returns the value associated with list_entry.  If the list is empty it
! returns LINKED_LIST_END_INTEGER.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer), intent(in) :: list_entry
integer :: linked_list_integer_value
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (trace_integer) write(outunit,*) "value in integer linked list entry is ",list_entry%value

linked_list_integer_value = list_entry%value

end function linked_list_integer_value

!        -------------------------------
function linked_list_integer_array_value(list_entry)
!        -------------------------------

!----------------------------------------------------
! Returns the value associated with list_entry.  If the list is empty it
! returns LINKED_LIST_END_INTEGER_ARRAY.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer_array), intent(in) :: list_entry
integer :: linked_list_integer_array_value(size(list_entry%value))
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (trace_integer_array) write(outunit,*) "value in integer array linked list entry is ",list_entry%value

linked_list_integer_array_value = list_entry%value

end function linked_list_integer_array_value

!        ------------------------
function linked_list_integer_next(list_entry)
!        ------------------------

!----------------------------------------------------
! Returns a pointer to the entry that follows list_entry in the linked list.
! If list_entry is the tail, the result is null, i.e. not associated.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer), intent(in) :: list_entry
type(linked_list_integer), pointer :: linked_list_integer_next
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (trace_integer) write(outunit,*) "getting next entry in integer linked list"

linked_list_integer_next => list_entry%next

end function linked_list_integer_next

!        ------------------------
function linked_list_integer_array_next(list_entry)
!        ------------------------

!----------------------------------------------------
! Returns a pointer to the entry that follows list_entry in the linked list.
! If list_entry is the tail, the result is null, i.e. not associated.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer_array), intent(in) :: list_entry
type(linked_list_integer_array), pointer :: linked_list_integer_array_next
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (trace_integer_array) write(outunit,*) "getting next entry in integer array linked list"

linked_list_integer_array_next => list_entry%next

end function linked_list_integer_array_next

!          -------------------------
subroutine linked_list_integer_print(list_head)
!          -------------------------

!----------------------------------------------------
! Prints the list of entries in a linked list
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer), intent(in), target :: list_head
!----------------------------------------------------
! Local variables:

type(linked_list_integer), pointer :: entry
!----------------------------------------------------
! Begin executable code

write(outunit,*) "Contents of linked list:"

if (list_head%value == LINKED_LIST_END_INTEGER) then
   write(outunit,*) "  linked list is empty"
else
   entry => list_head
   do while (associated(entry))
      write(outunit,*) "  ",entry%value
      entry => entry%next
   end do
endif

write(outunit,*) "  end of linked list"

end subroutine linked_list_integer_print

!          -------------------------------
subroutine linked_list_integer_array_print(list_head)
!          -------------------------------

!----------------------------------------------------
! Prints the list of entries in a linked list
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer_array), intent(in), target :: list_head
!----------------------------------------------------
! Local variables:

type(linked_list_integer_array), pointer :: entry
!----------------------------------------------------
! Begin executable code

write(outunit,*) "Contents of linked list:"

if (list_head%value .eqarray. LINKED_LIST_END_INTEGER_ARRAY) then
   write(outunit,*) "  linked list is empty"
else
   entry => list_head
   do while (associated(entry))
      write(outunit,*) "  ",entry%value
      entry => entry%next
   end do
endif

write(outunit,*) "  end of linked list"

end subroutine linked_list_integer_array_print

!===============================================================================
! Private routines

!                    -------------------------
recursive subroutine integer_destroy_recursive(entry)
!                    -------------------------

!----------------------------------------------------
! This routine recursively destroys the linked list starting at entry
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer), intent(inout) :: entry
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (associated(entry%next)) then
   call integer_destroy_recursive(entry%next)
   deallocate(entry%next)
endif

end subroutine integer_destroy_recursive

!                    -------------------------------
recursive subroutine integer_array_destroy_recursive(entry)
!                    -------------------------------

!----------------------------------------------------
! This routine recursively destroys the linked list starting at entry
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linked_list_integer_array), intent(inout) :: entry
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (associated(entry%next)) then
   call integer_array_destroy_recursive(entry%next)
   deallocate(entry%value,entry%next)
endif

end subroutine integer_array_destroy_recursive

!        -------------------
function integer_array_equal(array1,array2)
!        -------------------

!----------------------------------------------------
! This routine returns true if array1 and array2 have the same entries in
! any order.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: array1(:), array2(:)
logical :: integer_array_equal
!----------------------------------------------------
! Local variables:

integer :: arraycopy(size(array2))
integer :: i, j
!----------------------------------------------------
! Begin executable code

! Assume it is true and prove me wrong.

integer_array_equal = .true.

! They have to have the same size

if (size(array1) /= size(array2)) then

   integer_array_equal = .false.

else

! Make a copy of the second array

   arraycopy = array2

! Pass through the first array looking for each entry in the copy.  If found,
! change the entry in the copy to +infinity, and move on.  If not found, set
! the result to false and exit.

   outer: do i=1,size(array1)
      do j=1,size(arraycopy)
         if (array1(i) == arraycopy(j)) then
            arraycopy(j) = huge(0)
            cycle outer
         endif
      end do
      integer_array_equal = .false.
      exit
   end do outer

endif

end function integer_array_equal

!        ----------------------
function integer_array_notequal(array1,array2)
!        ----------------------

!----------------------------------------------------
! This routine returns true if array1 and array2 do not have the same entries in
! any order.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: array1(:), array2(:)
logical :: integer_array_notequal
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

integer_array_notequal = .not. integer_array_equal(array1,array2)

end function integer_array_notequal

end module linked_list
