!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Applied and Computational Mathematics Division                  !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

module make_linsys

!----------------------------------------------------
! This module contains routines to create and destroy the linear system
!
! communication tags in this module are of the form 11xx
!----------------------------------------------------

!----------------------------------------------------
! Other modules used are:

use global
use message_passing
use stopwatch
use hash_eq_mod
use sort_mod
use linsystype_mod
use gridtype_mod
use grid_util
use linsys_util
use quadrature_rules
use basis_functions
use basis_at_quad_pts
use evaluate
use laplace_elem_mat
use lapack_solve
use petsc_interf

!----------------------------------------------------

implicit none
private
public create_linear_system, &
       destroy_linear_system, &
       elemental_matrix         ! thread safe if assigning to ierr is

!----------------------------------------------------
! Non-module procedures used are:

!----------------------------------------------------

contains

!          --------------------
subroutine create_linear_system(linear_system,grid,procs,solver_cntl, &
                                io_cntl,still_sequential,notime)
!          --------------------

!----------------------------------------------------
! This routine creates the linear system.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linear_system
type(grid_type), intent(inout) :: grid
type(proc_info), intent(in) :: procs
type(solver_options), intent(in) :: solver_cntl
type(io_options), intent(in) :: io_cntl
logical, intent(in) :: still_sequential
logical, intent(in), optional :: notime

!----------------------------------------------------
! Local variables:

logical :: timeit
integer :: nelem_leaf, maxeq_per_elem, allocstat
integer, allocatable :: leaf_element(:)

!----------------------------------------------------
! Begin executable code

! If the linear system already exists, just update the right hand side

if (linear_system%linsys_exists .and. my_proc(procs) /= MASTER) then

! start timing

   if (present(notime)) then
      timeit = .not. notime
   else
      timeit = .true.
   endif
   if (timeit) then
      call reset_watch((/passemble,cpassemble/))
      call start_watch((/passemble,tassemble/))
   endif

! Dirichlet boundary conditions and true solution may have changed

   call update_dirichlet(grid,linear_system)

! Create a list of vertices for each floating region

   call make_floating_lists(grid,linear_system)

! Make a list of the leaf elements

   allocate(leaf_element(grid%nelem_leaf),stat=allocstat)
   if (allocstat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in create_linear_system",procs=procs)
      stop
   endif
   call make_leaf_list(grid,leaf_element,nelem_leaf, &
                       sem_simulation.and.linear_system%linsys_exists, &
                       no_floating=.true.)

! Compute the right hand side

   linear_system%rhs = 0.0_my_real
   call compute_matrix_rhs(linear_system,grid,procs,solver_cntl, &
                           linear_system%maxeq_per_elem,leaf_element, &
                           nelem_leaf,timeit,still_sequential)
   deallocate(leaf_element)

! Static condensation on the right hand side

   call static_condensation_rhs(linear_system,grid,procs)

! The residual has not been computed

   linear_system%relresid = -huge(0.0_my_real)

! Stop timing

   if (timeit) then
      call stop_watch((/passemble,tassemble/))
   endif

   return

endif

! MASTER doesn't have a linear system, but needs some things set

if (my_proc(procs) == MASTER) then
   linear_system%neq      = 0
   linear_system%neq_vert = 0
   linear_system%neq_edge = 0
   linear_system%neq_face = 0
   linear_system%neq_bubble = 0
   linear_system%neq_cond = 0
   linear_system%neq_full = 0
   nullify(linear_system%begin_row)
   linear_system%solver_niter = 0
   if (.not. linear_system%linsys_exists) linear_system%relresid = -huge(0.0_my_real)
   linear_system%linsys_exists = .true.
   linear_system%coarse_band_exists = .false.
   linear_system%lapack_symm_band_exists = .false.
   linear_system%lapack_gen_band_exists = .false.
   linear_system%petsc_matrix_exists = .false.
   nullify(linear_system%floating_next, linear_system%floating_head)
   return
endif

! Start timing the assembly process.

if (present(notime)) then
   timeit = .not. notime
else
   timeit = .true.
endif

if (timeit) then
   call reset_watch((/passemble,cpassemble/))
   call start_watch((/passemble,tassemble/))
endif

! Create a list of vertices for each floating region

call make_floating_lists(grid,linear_system)

! Create a list of all the leaf elements.

allocate(leaf_element(grid%nelem_leaf),stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in create_linear_system",procs=procs)
   return
endif
call make_leaf_list(grid,leaf_element,nelem_leaf,no_floating=.true.)

! Set the easy scalars.

linear_system%coarse_band_exists = .false.
linear_system%lapack_symm_band_exists = .false.
linear_system%lapack_gen_band_exists = .false.
linear_system%petsc_matrix_exists = .false.

linear_system%nlev_vert = phaml_global_max(procs,grid%nlev,1101)
linear_system%edge_level = linear_system%nlev_vert + 1
linear_system%face_level = linear_system%nlev_vert + 2
linear_system%bubble_level = linear_system%nlev_vert + 3
linear_system%beyond_last_level = linear_system%nlev_vert + 4
select case (global_element_kind)
case (TRIANGULAR_ELEMENT)
   linear_system%cond_level = linear_system%edge_level
case (TETRAHEDRAL_ELEMENT)
   linear_system%cond_level = linear_system%face_level
end select
linear_system%system_size = solver_cntl%system_size
linear_system%maxdeg = 1
linear_system%solver_niter = 0
linear_system%relresid = -huge(0.0_my_real)

! Determine number of vertex, edge, face and bubble equations and maximum
! number of equations related to any element.

call determine_neq(grid,linear_system,leaf_element,nelem_leaf,maxeq_per_elem)
linear_system%maxdeg = phaml_global_max(procs,linear_system%maxdeg,1100)
linear_system%maxeq_per_elem = maxeq_per_elem

! Create the hash table.

call hash_table_init(linear_system%eq_hash,linear_system%neq)

! Allocate space for linear_system.

call allocate_linsys(linear_system,solver_cntl,procs)

! Create the matrices for converting linear basis functions between
! the nodal basis and the h-hierarchical basis.

call create_basis_change(linear_system)

! Determine the order of the equations.  In the process, set begin_level,
! equation_type, iown, and equation gids, copy solution from the grid, and
! insert equations into the hash table.

call equation_order(linear_system,solver_cntl,grid,leaf_element,nelem_leaf)

! Create the column_index array.  This allocates column_index, and sets
! column_index, begin_row, end_row, end_row_linear, end_row_edge,
! end_row_face end_row_bubble

call create_column_index(linear_system,grid,procs,maxeq_per_elem, &
                         leaf_element,nelem_leaf)

! Allocate the matrix value arrays, and initialize them and the right hand
! side arrays to 0.0

call allocate_matrix(linear_system,solver_cntl,io_cntl,procs)

! Compute the matrix and right hand side.

call compute_matrix_rhs(linear_system,grid,procs,solver_cntl,maxeq_per_elem, &
                        leaf_element,nelem_leaf,timeit,still_sequential)

deallocate(leaf_element)

! Make need_r_others for HBMG multigrid

if (solver_cntl%solver == MG_SOLVER .or. &
    solver_cntl%preconditioner == MG_PRECONDITION) then
   call make_need_r_others(linear_system)
endif

! If the residual is to be computed while solving a boundary value problem
! (print_error_when is FREQUENTLY or TOO_MUCH) keep the stiffness by copying
! it to condensed.

if ((io_cntl%print_error_when == FREQUENTLY .or. &
     io_cntl%print_error_when == TOO_MUCH) .and. &
     solver_cntl%eq_type /= EIGENVALUE) then
   linear_system%condensed = linear_system%stiffness
   linear_system%rhs_cond = linear_system%rhs_nocond
endif

! Perform static condensation to eliminate the bubble bases for boundary
! value problems

if (solver_cntl%eq_type /= EIGENVALUE) then
   call static_condensation(linear_system,grid,procs)
else
   linear_system%matrix_val => linear_system%stiffness
   nullify(linear_system%elem_block)
endif

! The linear system now exists

linear_system%linsys_exists = .true.

! stop timing the assembly process

if (timeit) then
   call stop_watch((/passemble,tassemble/))
endif

end subroutine create_linear_system

!          ----------------
subroutine update_dirichlet(grid,linear_system)
!          ----------------

!----------------------------------------------------
! This routine updates the Dirichlet boundary conditions when the linear
! system already exists and we are just computing a new right hand side.
! It also updates the true solution, if known.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(linsys_type), intent(inout) :: linear_system
!----------------------------------------------------
! Local variables:

integer :: lev, elem, vert, edge, face, i, j, k, eq, objtype, brank, srank, lid
integer :: bctype(grid%system_size)
real(my_real) :: bccoef(grid%system_size,grid%system_size), &
                 bcrhs(grid%system_size)
logical(small_logical) :: visited_vert(grid%biggest_vert), &
                          visited_edge(grid%biggest_edge), &
                          visited_face(grid%biggest_face)

!----------------------------------------------------
! Begin executable code

! First update the solution components of the grid.

! Haven't visited any entities yet

visited_vert = .false.
visited_edge = .false.
visited_face = .false.

! For each leaf element ...

do lev=1,grid%nlev
   elem = grid%head_level_elem(lev)
   do while (elem /= END_OF_LIST)
      if (grid%element(elem)%isleaf) then

! Set the element true solution, if it is known.

         if (grid%have_true) then
            if (grid%element(elem)%degree > 2) then
               do j=1,grid%system_size
                  call elem_exact(grid,elem,j,"t")
               end do
            endif 
         endif

! For each unvisited vertex ...

         do i=1,VERTICES_PER_ELEMENT
            vert = grid%element(elem)%vertex(i)
            if (visited_vert(vert)) cycle
            visited_vert(vert) = .true.

! Set the true solution, if it is known.

            if (grid%have_true) then
               do k=1,max(1,grid%num_eval)
                  do j=1,grid%system_size
                     call set_vert_exact(grid,vert,j,k, &
                               my_trues(grid%vertex(vert)%coord%x, &
                                        grid%vertex(vert)%coord%y, &
                                        zcoord(grid%vertex(vert)%coord),j,k))
                  end do
               end do
            endif

! If it is a boundary vertex, get the boundary conditions and set the
! Dirichlet components

            if (grid%vertex_type(vert,1) /= INTERIOR) then
               secret_elem = elem
               call my_bconds(grid%vertex(vert)%coord%x, &
                              grid%vertex(vert)%coord%y, &
                              zcoord(grid%vertex(vert)%coord), &
                              grid%vertex(vert)%bmark,bctype,bccoef,bcrhs)
               do j=1,grid%system_size
                  if (grid%vertex_type(vert,j) == DIRICHLET .or. &
                      grid%vertex_type(vert,j) == INTERIOR_DIRICHLET .or. &
                      grid%vertex_type(vert,j) == PERIODIC_SLAVE_DIR .or. &
                      grid%vertex_type(vert,j) == PERIODIC_MASTER_DIR) then
                     call set_vert_soln(grid,vert,j,colon,bcrhs(j))
                  endif
               end do
            endif
         end do

! Only have to do edges if the maximum degree is more than 1

         if (linear_system%maxdeg > 1) then

! For each unvisited edge ...

            do i=1,EDGES_PER_ELEMENT
               edge = grid%element(elem)%edge(i)
               if (visited_edge(edge)) cycle
               visited_edge(edge) = .true.

! Only need to do edges with degree more than 1

               if (grid%edge(edge)%degree > 1) then

! Set the true solution, if it is known.

                  if (grid%have_true) then
                     do j=1,grid%system_size
                        call elem_exact(grid,edge,j,"t")
                     end do
                  endif

! If it is a boundary edge, set the edge solutions for each Dirichlet component.

                  if (grid%edge_type(edge,1) /= INTERIOR) then
                     do j=1,grid%system_size
                        if (grid%edge_type(edge,j) == DIRICHLET .or. &
                            grid%edge_type(edge,j) == INTERIOR_DIRICHLET .or. &
                            grid%edge_type(edge,j) == PERIODIC_SLAVE_DIR .or. &
                            grid%edge_type(edge,j) == PERIODIC_MASTER_DIR) then
                           call edge_exact(grid,edge,j,"d")
                        endif
                     end do
                  endif
               endif
            end do

         end if

! Only have to do faces if the maximum degree is more than 2

         if (linear_system%maxdeg > 2) then

! For each unvisited face ...

            do i=1,FACES_PER_ELEMENT
               face = grid%element(elem)%face(i)
               if (visited_face(face)) cycle
               visited_face(face) = .true.

! Only need to do faces with degree more than 2

               if (grid%face(face)%degree > 2) then

! Set the true solution, if it is known.

                  if (grid%have_true) then
                     do j=1,grid%system_size
                        call face_exact(grid,face,j,"t")
                     end do
                  endif

! If it is a boundary face, set the face solutions for each Dirichlet component.

                  if (grid%face_type(face,1) /= INTERIOR) then
                     do j=1,grid%system_size
                        if (grid%face_type(face,j) == DIRICHLET .or. &
                            grid%face_type(face,j) == INTERIOR_DIRICHLET .or. &
                            grid%face_type(face,j) == PERIODIC_SLAVE_DIR .or. &
                            grid%face_type(face,j) == PERIODIC_MASTER_DIR) then
                           call face_exact(grid,face,j,"d")
                        endif
                     end do
                  endif
               endif
            end do

         endif

! Next element

      endif ! isleaf
      elem = grid%element(elem)%next
   end do
end do

! Then copy the Dirichlet boundary conditions to the linear system right
! hand side and solution.

! For each Dirichlet equation ...

do eq=1,linear_system%neq
   if (linear_system%equation_type(eq) /= DIRICHLET) cycle

! Find the corresponding grid entity

   call eq_to_grid(linear_system,linear_system%gid(eq),objtype,brank,srank, &
                   lid,grid)

! Copy the grid solution to the linear system right hand side and solution

   select case (objtype)
   case (VERTEX_ID)
      linear_system%rhs(eq) = get_vert_soln(grid,lid,srank,1)
   case (EDGE_ID)
      linear_system%rhs(eq) = get_edge_soln(grid,lid,brank,srank,1)
   case (FACE_ID)
      linear_system%rhs(eq) = get_face_soln(grid,lid,brank,srank,1)
   case (ELEMENT_ID)
      linear_system%rhs(eq) = get_elem_soln(grid,lid,brank,srank,1)
   end select
   linear_system%solution(eq) = linear_system%rhs(eq)

end do

end subroutine update_dirichlet

!          -------------------
subroutine make_floating_lists(grid,linear_system)
!          -------------------

!----------------------------------------------------
! This routine creates linked lists of the vertices associated with each
! floating region, and counts the number of boundary vertices in each
! floating region.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(linsys_type), intent(inout) :: linear_system
!----------------------------------------------------
! Local variables:

logical(small_logical) :: visited_vert(grid%biggest_vert)
integer :: i, lev, elem, region, ivert, vert, previous, oldhead, oldvertnext, &
           astat
integer, pointer :: elements(:)
!----------------------------------------------------
! Begin executable code

! If no floating regions, nothing to do.

if (.not. associated(grid%floating_tags)) then
   nullify(linear_system%floating_next, linear_system%floating_head)
   return
endif

! Allocate linked list of vertices in floating regions

allocate(linear_system%floating_next(grid%biggest_vert), &
         linear_system%floating_head(size(grid%floating_tags)), stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in make_floating_lists")
   stop
endif
linear_system%floating_head = END_OF_LIST

! Keep track of which vertices have already been seen

visited_vert = .false.

! For each leaf element in a floating region ...

do lev=1,grid%nlev
   elem = grid%head_level_elem(lev)
   do while (elem /= END_OF_LIST)
      if (grid%element(elem)%isleaf .and. &
          grid%element_type(elem,1) == FLOATING) then
         region = -2

! For each vertex that has not already been seen

         do ivert=1,VERTICES_PER_ELEMENT
            vert = grid%element(elem)%vertex(ivert)
            if (visited_vert(vert)) cycle
            visited_vert(vert) = .true.

! Get the index of the floating region that contains elem, if not already done

            if (region == -2) then
               region = get_elem_floating_region(grid,elem)
               if (region == -1) then
                  ierr = PHAML_INTERNAL_ERROR
                  call fatal("make_floating_lists: looking for floating region of nonfloating element")
                  stop
               endif
            endif

! Add vertex to the list

            linear_system%floating_next(vert) = &
               linear_system%floating_head(region)
            linear_system%floating_head(region) = vert

         end do ! next vertex
      endif ! floating leaf
      elem = grid%element(elem)%next
   end do ! next element
end do ! next level

! Now make sure the head of each list is on the boundary of the floating region.

do i=1,size(grid%floating_tags)
   vert = linear_system%floating_head(i)
   do
      call get_vertex_elements(grid,vert,elements)
      if (any(grid%element_type(elements,1) /= FLOATING)) then
         deallocate(elements)
         exit
      endif
      deallocate(elements)
      previous = vert
      vert = linear_system%floating_next(vert)
      if (vert == END_OF_LIST) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("make_floating_lists: didn't find a floating vertex on the boundary")
         stop
      endif
   end do
   if (vert /= linear_system%floating_head(i)) then
      oldhead = linear_system%floating_head(i)
      oldvertnext = linear_system%floating_next(vert)
      linear_system%floating_head(i) = vert
      if (linear_system%floating_next(oldhead) == vert) then
         linear_system%floating_next(vert) = oldhead
      else
         linear_system%floating_next(vert)=linear_system%floating_next(oldhead)
         linear_system%floating_next(previous) = oldhead
      endif
      linear_system%floating_next(oldhead) = oldvertnext
   endif
end do
end subroutine make_floating_lists

!          --------------
subroutine make_leaf_list(grid,leaf_element,nelem,skip_zero_charge,no_floating)
!          --------------

!----------------------------------------------------
! This routine makes a list of the leaf elements
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
integer, intent(inout) :: leaf_element(:), nelem
logical, intent(in), optional :: skip_zero_charge, no_floating

!----------------------------------------------------
! Local variables:

integer :: elem, lev
logical :: loc_skip_zero_charge, loc_no_floating

!----------------------------------------------------
! Begin executable code

! for SEM simulation with linear system already existing, only do rhs
! update for leaves with a nonzero charge or a Neumann boundary condition

if (present(skip_zero_charge)) then
   loc_skip_zero_charge = skip_zero_charge
else
   loc_skip_zero_charge = .false.
endif

! Don't include elements that are in a floating region.

if (present(no_floating)) then
   loc_no_floating = no_floating
else
   loc_no_floating = .false.
endif

! go through the refinement tree making a list of leaf elements

nelem = 0
do lev=1,grid%nlev
   elem = grid%head_level_elem(lev)
   do while (elem /= END_OF_LIST)
      if (grid%element(elem)%isleaf) then
         if (loc_no_floating) then
            if (get_elem_floating_region(grid,elem) /= -1) then
               elem = grid%element(elem)%next
               cycle
            endif
         endif
         if (loc_skip_zero_charge) then
            if (grid%element(elem)%tags(chargetag) /= 0 .or. &
                any(grid%face_type(grid%element(elem)%face,:)==NATURAL)) then
               nelem = nelem + 1
               leaf_element(nelem) = elem
            endif
         else
            nelem = nelem + 1
            leaf_element(nelem) = elem
         endif
      endif
      elem = grid%element(elem)%next
   end do
end do
if (nelem /= grid%nelem_leaf .and. .not. loc_skip_zero_charge .and. &
    .not. loc_no_floating) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("number of leaf elements found is not equal to number of leaf elements in grid%nelem_leaf in make_leaf_list", &
              intlist=(/nelem,grid%nelem_leaf/))
   stop
endif

end subroutine make_leaf_list

!          -------------
subroutine determine_neq(grid,linear_system,leaf_element,nelem_leaf, &
                         maxeq_per_elem)
!          -------------

!----------------------------------------------------
! This routine determines the number of vertex, edge, face and bubble equations
! and maximum number of equations related to any element
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linear_system
type(grid_type), intent(in) :: grid
integer, intent(in) :: leaf_element(:), nelem_leaf
integer, intent(out) :: maxeq_per_elem
!----------------------------------------------------
! Local variables:

logical(small_logical) :: visited_vert(grid%biggest_vert), &
                          visited_edge(grid%biggest_edge), &
                          visited_face(grid%biggest_face)
integer :: i, j, k, elem, face, edge, vert, degree, eq_this_elem, lev, &
           syssize, neqvert, neqedge, neqface, neqbubble
!----------------------------------------------------
! Begin executable code

syssize = linear_system%system_size
visited_edge = .false.
visited_face = .false.
neqvert = grid%nvert
neqedge = 0
neqface = 0
neqbubble = 0
maxeq_per_elem = 0
do j=1,nelem_leaf
   elem = leaf_element(j)
   eq_this_elem = VERTICES_PER_ELEMENT
   do i=1,EDGES_PER_ELEMENT
      edge = grid%element(elem)%edge(i)
      degree = grid%edge(edge)%degree
      if (degree > 1) then
         eq_this_elem = eq_this_elem + degree-1
         if (visited_edge(edge)) cycle
         neqedge = neqedge + degree - 1
         visited_edge(edge) = .true.
      endif
   end do
   do i=1,FACES_PER_ELEMENT
      face = grid%element(elem)%face(i)
      degree = grid%face(face)%degree
      if (degree > 2) then
         eq_this_elem = eq_this_elem + face_dof(degree)
         if (visited_face(face)) cycle
         neqface = neqface + face_dof(degree)
         visited_face(face) = .true.
      endif
   end do
   degree = grid%element(elem)%degree
   linear_system%maxdeg = max(degree,linear_system%maxdeg)
   if (degree > 2) then
      eq_this_elem = eq_this_elem + element_dof(degree)
      neqbubble = neqbubble + element_dof(degree)
   endif
   maxeq_per_elem = max(maxeq_per_elem,eq_this_elem)
end do
neqvert = neqvert*syssize
neqedge = neqedge*syssize
neqface = neqface*syssize
neqbubble = neqbubble*syssize

! remove slave periodic entities from equation count

do lev=1,grid%nlev
   elem = grid%head_level_elem(lev)
   do while (elem /= END_OF_LIST)
      elem = grid%element(elem)%next
   end do
end do
visited_vert = .false.
visited_edge = .false.
visited_face = .false.
do k=1,nelem_leaf
   elem = leaf_element(k)
   do i=1,VERTICES_PER_ELEMENT
      vert = grid%element(elem)%vertex(i)
      if (visited_vert(vert)) cycle
      visited_vert(vert) = .true.
      do j=1,syssize
         if (is_periodic_vert_slave(vert,grid,j)) then
            neqvert = neqvert - 1
         endif
      end do
   end do
   do i=1,EDGES_PER_ELEMENT
      edge = grid%element(elem)%edge(i)
      if (visited_edge(edge)) cycle
      visited_edge(edge) = .true.
      if (grid%edge(edge)%degree < 2) cycle
      do j=1,syssize
         if (is_periodic_edge_slave(edge,grid,j)) then
            neqedge = neqedge - (grid%edge(edge)%degree-1)
         endif
      end do
   end do
   do i=1,FACES_PER_ELEMENT
      face = grid%element(elem)%face(i)
      if (visited_face(face)) cycle
      visited_face(face) = .true.
      if (grid%face(face)%degree < 3) cycle
      do j=1,syssize
         if (is_periodic_face_slave(face,grid,j)) then
            neqface = neqface - face_dof(grid%face(face)%degree)
         endif
      end do
   end do
end do

! Replace the number of vertices (equations) in floating regions by 1 for
! each region.

if (associated(grid%floating_tags)) then
   neqvert = neqvert + size(grid%floating_tags)
   do i=1,size(grid%floating_tags)
      vert = linear_system%floating_head(i)
      do while (vert /= END_OF_LIST)
         neqvert = neqvert - 1
         vert = linear_system%floating_next(vert)
      end do
   end do
endif

linear_system%neq_cond = neqvert + neqedge + neqface
linear_system%neq_full = neqvert + neqedge + neqface + neqbubble
linear_system%neq = neqvert + neqedge + neqface + neqbubble
linear_system%neq_vert = neqvert
linear_system%neq_edge = neqedge
linear_system%neq_face = neqface
linear_system%neq_bubble = neqbubble

end subroutine determine_neq

!          ---------------
subroutine allocate_linsys(linear_system,solver_cntl,procs)
!          ---------------

!----------------------------------------------------
! This routine allocates memory in linear_system
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linear_system
type(solver_options), intent(in) :: solver_cntl
type(proc_info), intent(in) :: procs
!----------------------------------------------------
! Local variables:

integer :: neq, syssize, allocstat
!----------------------------------------------------
! Begin executable code

neq = linear_system%neq
syssize = linear_system%system_size

! allocate space based on neq, nlev, etc.
! Any changes to this list should be matched in the deallocate statement
! in destroy_linear_system

allocate(linear_system%begin_row(neq+1), &
         linear_system%end_row_linear(neq+1), &
         linear_system%end_row_edge(neq+1), &
         linear_system%end_row_bubble(neq+1), &
         linear_system%begin_level(linear_system%beyond_last_level), &
         linear_system%rhs_nocond(neq), &
         linear_system%r_mine(neq), &
         linear_system%r_others(neq), &
         linear_system%need_r_others(neq), &
         linear_system%iown(neq), &
         linear_system%solution(0:neq), &
         linear_system%equation_type(neq), &
         linear_system%gid(neq), &
         linear_system%hold_dirich(neq), &
         linear_system%s_int(5*syssize,5*syssize), &
         linear_system%s_int_inv(5*syssize,5*syssize), &
         linear_system%s_bnd(4*syssize,4*syssize), &
         linear_system%s_bnd_inv(4*syssize,4*syssize), &
         stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in allocate_linsys",procs=procs)
   return
endif
select case (global_element_kind)
case (TRIANGULAR_ELEMENT)
   nullify(linear_system%end_row_face)
case (TETRAHEDRAL_ELEMENT)
   allocate(linear_system%end_row_face(neq+1),stat=allocstat)
   if (allocstat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in allocate_linsys",procs=procs)
      return
   endif
end select
linear_system%end_row => linear_system%end_row_bubble
linear_system%rhs => linear_system%rhs_nocond
if (solver_cntl%num_eval > 1) then
   allocate(linear_system%evecs(neq,solver_cntl%num_eval-1),stat=allocstat)
   if (allocstat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in allocate_linsys",procs=procs)
      return
   endif
else
   nullify(linear_system%evecs)
endif
nullify(linear_system%matrix_val,linear_system%column_index, &
        linear_system%stiffness,linear_system%mass)

end subroutine allocate_linsys

!          -------------------
subroutine create_basis_change(linear_system)
!          -------------------

!----------------------------------------------------
! This routine creates the matrices for converting linear basis functions
! between ! the nodal basis and the h-hierarchical basis
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linear_system
!----------------------------------------------------
! Local variables:

integer :: i, syssize
!----------------------------------------------------
! Begin executable code

syssize = linear_system%system_size

linear_system%s_int = 0.0_my_real
linear_system%s_int_inv = 0.0_my_real
linear_system%s_bnd = 0.0_my_real
linear_system%s_bnd_inv = 0.0_my_real

do i=1,5*syssize
   linear_system%s_int(i,i) = 1.0_my_real
   linear_system%s_int_inv(i,i) = 1.0_my_real
end do
do i=1,4*syssize
   linear_system%s_bnd(i,i) = 1.0_my_real
   linear_system%s_bnd_inv(i,i) = 1.0_my_real
end do

do i=1,syssize
   linear_system%s_int(4*syssize+i,i) = 0.5_my_real
   linear_system%s_int(4*syssize+i,i+syssize) = 0.5_my_real
   linear_system%s_int_inv(4*syssize+i,i) = -0.5_my_real
   linear_system%s_int_inv(4*syssize+i,i+syssize) = -0.5_my_real
   linear_system%s_bnd(3*syssize+i,i) = 0.5_my_real
   linear_system%s_bnd(3*syssize+i,i+syssize) = 0.5_my_real
   linear_system%s_bnd_inv(3*syssize+i,i) = -0.5_my_real
   linear_system%s_bnd_inv(3*syssize+i,i+syssize) = -0.5_my_real
end do

end subroutine create_basis_change

!          --------------
subroutine equation_order(linear_system,solver_control,grid,leaf_element, &
                          nelem_leaf)
!          --------------

!----------------------------------------------------
! This routine orders the equations with vertex bases (linear bases) first,
! ordered by level, and then edge bases ordered by degree, and then face bases
! ordered by degree, and then bubble bases ordered by element.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linear_system
type(solver_options), intent(in) :: solver_control
type(grid_type), intent(in) :: grid
integer, intent(in) :: leaf_element(:), nelem_leaf
!----------------------------------------------------
! Local variables:

logical(small_logical) :: visited_vert(grid%biggest_vert), &
                          visited_edge(grid%biggest_edge), &
                          visited_face(grid%biggest_face)
integer :: i, j, k, l, m, eq, lev, vert, edge, face, elem, degree, syssize, &
           nev, brank, ivert, astat, lid, region
integer, allocatable :: eqa(:)
type(hash_key_eq) :: gid
!----------------------------------------------------
! Begin executable code

syssize = linear_system%system_size
nev = solver_control%num_eval
linear_system%solution(0) = 0.0_my_real

! Mark vertices in floating regions, except the head of the list, as already
! visited so they are not included.

visited_vert = .false.
if (associated(linear_system%floating_head)) then
   do region=1,size(linear_system%floating_head)
      vert = linear_system%floating_head(region)
      vert = linear_system%floating_next(vert)
      do while (vert /= END_OF_LIST)
         visited_vert(vert) = .true.
         vert = linear_system%floating_next(vert)
      end do
   end do
endif

! First pass through the vertices to determine the linear-basis equation order,
! global IDs, initial solution, equation type and ownership, and the start of
! equations for each level.  For triangles, put the nonpeak vertices before
! the peaks so that ancestors of level 2 equations come before parents.  Do
! this by starting with a fake lev=0 that picks up the nonpeaks.

eq = 1
do lev=0,grid%nlev
   if (lev==0 .and. global_element_kind == TETRAHEDRAL_ELEMENT) cycle
   if (lev==0) then
      linear_system%begin_level(1) = eq
      elem = grid%head_level_elem(1)
   else
      if (lev /= 1 .or. global_element_kind == TETRAHEDRAL_ELEMENT) then
         linear_system%begin_level(lev) = eq
      endif
      elem = grid%head_level_elem(lev)
   endif
   do while (elem /= END_OF_LIST)
      do ivert=1,VERTICES_PER_ELEMENT
         if (lev==0 .and. ivert==3) cycle
         vert = grid%element(elem)%vertex(ivert)
         if (visited_vert(vert)) cycle
         visited_vert(vert) = .true.
         if (grid%element_type(elem,1) == FLOATING) then
            region = get_elem_floating_region(grid,elem)
            if (linear_system%floating_head(region) /= vert) cycle
         endif
         do j=1,syssize
            if (is_periodic_vert_slave(vert,grid,j)) cycle
            call grid_to_eq(grid,linear_system,VERTEX_ID,1,j, &
                           grid%vertex(vert)%gid,eqn_gid=linear_system%gid(eq))
            call hash_insert(linear_system%gid(eq),eq,linear_system%eq_hash)
            linear_system%solution(eq) = get_vert_soln(grid,vert,j,1)
            if (nev > 1) then
               linear_system%evecs(eq,1:nev-1) = get_vert_soln(grid,vert,j, &
                                                               (/2,nev/))
            endif
            select case (grid%vertex_type(vert,j))
            case (INTERIOR_DIRICHLET)
               linear_system%equation_type(eq) = DIRICHLET
            case (PERIODIC_MASTER_DIR)
               linear_system%equation_type(eq) = DIRICHLET
            case (PERIODIC_MASTER_NAT)
               linear_system%equation_type(eq) = NATURAL
            case (PERIODIC_MASTER_MIX)
               linear_system%equation_type(eq) = MIXED
            case default
               linear_system%equation_type(eq) = grid%vertex_type(vert,j)
            end select
            linear_system%iown(eq) = &
               grid%element(grid%vertex(vert)%assoc_elem)%iown
            eq = eq + 1
         end do
      end do
      elem = grid%element(elem)%next
   end do
end do

do lev=grid%nlev+1,linear_system%nlev_vert
   linear_system%begin_level(lev) = eq
end do

if (associated(grid%floating_tags)) then
   do region=1,size(grid%floating_tags)
      vert = linear_system%floating_head(region)
      call grid_to_eq(grid,linear_system,VERTEX_ID,1,1, &
                      grid%vertex(vert)%gid,lid)
      vert = linear_system%floating_next(vert)
      do while (vert /= END_OF_LIST)
         call grid_to_eq(grid,linear_system,VERTEX_ID,1,1, &
                         grid%vertex(vert)%gid,eqn_gid=gid)
         call hash_insert(gid,lid,linear_system%eq_hash)
         vert = linear_system%floating_next(vert)
      end do
   end do
endif

if (eq-1 /= linear_system%neq_vert) then
   call warning("number of vertex equations created is not equal to neqvert")
endif

! Next pass through the edges to set the same components of the edge basis
! equations.  First go through the edges to determine how many equations there
! are for each degree.  Count the equations with degree in
! edge_begin_degree(degree+1) and then do a running sum to determine the
! first equation of each degree

linear_system%begin_level(linear_system%edge_level) = eq

allocate(linear_system%edge_begin_degree(linear_system%maxdeg+1), &
         eqa(linear_system%maxdeg), stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in equation_order")
   return
endif

linear_system%edge_begin_degree = 0
visited_edge = .false.
do l=1,nelem_leaf
   elem = leaf_element(l)
   do i=1,EDGES_PER_ELEMENT
      edge = grid%element(elem)%edge(i)
      if (visited_edge(edge)) cycle
      visited_edge(edge) = .true.
      degree = grid%edge(edge)%degree
      if (degree < 2) cycle
      do k=2,degree
         do j=1,syssize
            if (is_periodic_edge_slave(edge,grid,j)) cycle
            linear_system%edge_begin_degree(k+1) = &
                    linear_system%edge_begin_degree(k+1) + 1
         end do
      end do
   end do
end do

linear_system%edge_begin_degree(1) = eq
linear_system%edge_begin_degree(2) = eq
do i=3,linear_system%maxdeg+1
   linear_system%edge_begin_degree(i) = linear_system%edge_begin_degree(i) + &
                                        linear_system%edge_begin_degree(i-1)
end do
eqa = linear_system%edge_begin_degree(1:linear_system%maxdeg)

visited_edge = .false.
do l=1,nelem_leaf
   elem = leaf_element(l)
   do i=1,EDGES_PER_ELEMENT
      edge = grid%element(elem)%edge(i)
      if (visited_edge(edge)) cycle
      visited_edge(edge) = .true.
      degree = grid%edge(edge)%degree
      if (degree < 2) cycle
      do k=1,degree-1
         do j=1,syssize
            if (is_periodic_edge_slave(edge,grid,j)) cycle
            call grid_to_eq(grid,linear_system,EDGE_ID,k,j, &
                          grid%edge(edge)%gid, &
                          eqn_gid=linear_system%gid(eqa(k+1)))
            call hash_insert(linear_system%gid(eqa(k+1)),eqa(k+1), &
                             linear_system%eq_hash)
            linear_system%solution(eqa(k+1)) = get_edge_soln(grid,edge,k,j,1)
            if (nev > 1) then
               linear_system%evecs(eqa(k+1),1:nev-1) = &
                                          get_edge_soln(grid,edge,k,j,(/2,nev/))
            endif
            select case (grid%edge_type(edge,j))
            case (INTERIOR_DIRICHLET)
               linear_system%equation_type(eqa(k+1)) = DIRICHLET
            case (PERIODIC_MASTER_DIR)
               linear_system%equation_type(eqa(k+1)) = DIRICHLET
            case (PERIODIC_MASTER_NAT)
               linear_system%equation_type(eqa(k+1)) = NATURAL
            case (PERIODIC_MASTER_MIX)
               linear_system%equation_type(eqa(k+1)) = MIXED
            case default
               linear_system%equation_type(eqa(k+1)) = grid%edge_type(edge,j)
            end select
            linear_system%iown(eqa(k+1)) = &
               grid%element(grid%edge(edge)%assoc_elem)%iown
            eqa(k+1) = eqa(k+1) + 1
         end do
      end do
   end do
end do

do i=2,linear_system%maxdeg
   if (eqa(i) /= linear_system%edge_begin_degree(i+1)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("equation_order: mismatch in number of edge equations")
      stop
   endif
end do

eq = linear_system%edge_begin_degree(linear_system%maxdeg+1)

if (eq-1 /= linear_system%neq_vert + linear_system%neq_edge) then
   call warning("number of edge equations is not equal to neqedge")
endif

! Same for face basis equations.

linear_system%begin_level(linear_system%face_level) = eq

allocate(linear_system%face_begin_degree(linear_system%maxdeg+1), stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in equation_order")
   return
endif

if (global_element_kind == TETRAHEDRAL_ELEMENT) then

   linear_system%face_begin_degree = 0
   visited_face = .false.
   do l=1,nelem_leaf
      elem = leaf_element(l)
      do i=1,FACES_PER_ELEMENT
         face = grid%element(elem)%face(i)
         if (visited_face(face)) cycle
         visited_face(face) = .true.
         degree = grid%face(face)%degree
         if (degree < 3) cycle
         do k=3,degree
            do j=1,syssize
               if (is_periodic_face_slave(face,grid,j)) cycle
               linear_system%face_begin_degree(k+1) = &
                       linear_system%face_begin_degree(k+1) + k-2
            end do
         end do
      end do
   end do

   linear_system%face_begin_degree(1) = eq
   linear_system%face_begin_degree(2) = eq
   if (linear_system%maxdeg > 1) linear_system%face_begin_degree(3) = eq
   do i=4,linear_system%maxdeg+1
      linear_system%face_begin_degree(i) = linear_system%face_begin_degree(i)+ &
                                           linear_system%face_begin_degree(i-1)
   end do
   eqa = linear_system%face_begin_degree(1:linear_system%maxdeg)

   visited_face = .false.
   do l=1,nelem_leaf
      elem = leaf_element(l)
      do i=1,FACES_PER_ELEMENT
         face = grid%element(elem)%face(i)
         if (visited_face(face)) cycle
         visited_face(face) = .true.
         degree = grid%face(face)%degree
         if (degree < 3) cycle
         brank = 0
         do k=3,degree
            do j=1,syssize
               if (is_periodic_face_slave(face,grid,j)) cycle
               do m=1,k-2
                  if (j==1) brank = brank+1
                  call grid_to_eq(grid,linear_system,FACE_ID,brank,j, &
                                  grid%face(face)%gid, &
                                  eqn_gid=linear_system%gid(eqa(k)))
                  call hash_insert(linear_system%gid(eqa(k)),eqa(k), &
                                   linear_system%eq_hash)
                  linear_system%solution(eqa(k)) = &
                                              get_face_soln(grid,face,brank,j,1)
                  if (nev > 1) then
                     linear_system%evecs(eqa(k),1:nev-1) = &
                                      get_face_soln(grid,face,brank,j,(/2,nev/))
                  endif
                  select case (grid%face_type(face,j))
                  case (INTERIOR_DIRICHLET)
                     linear_system%equation_type(eqa(k)) = DIRICHLET
                  case (PERIODIC_MASTER_DIR)
                     linear_system%equation_type(eqa(k)) = DIRICHLET
                  case (PERIODIC_MASTER_NAT)
                     linear_system%equation_type(eqa(k)) = NATURAL
                  case (PERIODIC_MASTER_MIX)
                     linear_system%equation_type(eqa(k)) = MIXED
                  case default
                     linear_system%equation_type(eqa(k)) =grid%face_type(face,j)
                  end select
                  linear_system%iown(eqa(k)) = &
                        grid%element(grid%face(face)%assoc_elem)%iown
                  eqa(k) = eqa(k) + 1
               end do
            end do
         end do
      end do
   end do

   do i=3,linear_system%maxdeg
      if (eqa(i) /= linear_system%face_begin_degree(i+1)) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("equation_order: mismatch in number of face equations")
         stop
      endif
   end do

   eq = linear_system%face_begin_degree(linear_system%maxdeg+1)

   if (eq-1 /= linear_system%neq_vert + linear_system%neq_edge + &
               linear_system%neq_face) then
      call warning("number of face equations is not equal to neqface")
   endif

else
   linear_system%face_begin_degree = eq
endif

! Then pass through the elements setting the same components of the
! bubble basis equations.

linear_system%begin_level(linear_system%bubble_level) = eq

do l=1,nelem_leaf
   elem = leaf_element(l)
   degree = grid%element(elem)%degree
   if (degree > 2) then
      do k=1,element_dof(degree)
         do j=1,syssize
            call grid_to_eq(grid,linear_system,ELEMENT_ID,k,j, &
                 grid%element(elem)%gid,eqn_gid=linear_system%gid(eq+j-1))
            call hash_insert(linear_system%gid(eq+j-1),eq+j-1, &
                             linear_system%eq_hash)
            if (grid%element_type(elem,j) == INTERIOR_DIRICHLET) then
               linear_system%equation_type(eq+j-1) = DIRICHLET
            else
               linear_system%equation_type(eq+j-1) = INTERIOR
            endif
         end do
         linear_system%solution(eq:eq+syssize-1) = &
                                              get_elem_soln(grid,elem,k,colon,1)
         if (nev > 1) then
            linear_system%evecs(eq:eq+syssize-1,1:nev-1) = &
                                 get_elem_soln(grid,elem,k,colon,(/2,nev/))
         endif
         linear_system%iown(eq:eq+syssize-1) = grid%element(elem)%iown
         eq = eq + syssize
      end do
   endif
end do

if (eq-1 /= linear_system%neq_vert + linear_system%neq_edge + &
            linear_system%neq_face + linear_system%neq_bubble) then
   call warning("number of bubble equations is not equal to neqbubble")
endif

linear_system%begin_level(linear_system%beyond_last_level) = eq

deallocate(eqa)

end subroutine equation_order

!          -------------------
subroutine old_create_column_index(linear_system,grid,procs,maxeq_per_elem, &
                                   leaf_element,nelem_leaf)
!          -------------------

!----------------------------------------------------
! This routine creates column_index and related arrays.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linear_system
type(grid_type), intent(inout) :: grid
type(proc_info), intent(in) :: procs
integer, intent(in) :: maxeq_per_elem, leaf_element(:), nelem_leaf
!----------------------------------------------------
! Local variables:

logical :: visited_vert(grid%biggest_vert), visited_edge(grid%biggest_edge), &
           visited_face(grid%biggest_face)
integer :: i, j, k, l, m, lev, vert, edge, face, elem, eq, eq1, neq, nentry, &
           degree, indx, syssize, objtype2, brank, srank, allocstat, &
           region, max_entries_in_row, numobj, numdof, iobj, ivert
integer :: eqlist(maxeq_per_elem)
integer, allocatable :: adjacencies(:,:), objtype(:), objid(:), first_indx(:), &
                        numadj(:)
!----------------------------------------------------
! Begin executable code

neq = linear_system%neq
syssize = linear_system%system_size

! Determine the maximum number of entries in any row of the matrix

max_entries_in_row = find_max_entries_in_row(grid,leaf_element,nelem_leaf)

! Pass through the elements to determine which bases have
! overlapping support.  For each equation build a temporary list
! of equations from this pass, and then create the column_index array

! create the lists

allocate(adjacencies(max_entries_in_row,neq),numadj(neq),stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation for adjacencies failed",procs=procs)
   return
endif

numadj = 0
nentry = 0

! Pass through the elements building lists and counting entries.
! The list only contains the first equation of each block if system_size > 1.

!$omp parallel do schedule(dynamic) &
!$omp  default(shared) &
!$omp  private(i,eq,j,eqlist,brank,degree)

do i=1,nelem_leaf
   eq = 1
   do j=1,VERTICES_PER_ELEMENT
      call grid_to_eq(grid,linear_system,VERTEX_ID,1,1, &
          grid%vertex(grid%element(leaf_element(i))%vertex(j))%gid,eqlist(eq))
      eq = eq + 1
   end do
   do j=1,EDGES_PER_ELEMENT
      do brank = 1,grid%edge(grid%element(leaf_element(i))%edge(j))%degree-1
         call grid_to_eq(grid,linear_system,EDGE_ID,brank,1, &
               grid%edge(grid%element(leaf_element(i))%edge(j))%gid,eqlist(eq))
         eq = eq + 1
      end do
   end do
   do j=1,FACES_PER_ELEMENT
      degree = grid%face(grid%element(leaf_element(i))%face(j))%degree
      do brank = 1,face_dof(degree)
         call grid_to_eq(grid,linear_system,FACE_ID,brank,1, &
               grid%face(grid%element(leaf_element(i))%face(j))%gid,eqlist(eq))
         eq = eq + 1
      end do
   end do
   degree = grid%element(leaf_element(i))%degree
   do j=1,element_dof(degree)
      call grid_to_eq(grid,linear_system,ELEMENT_ID,j,1, &
                      grid%element(leaf_element(i))%gid,eqlist(eq))
      eq = eq + 1
   end do
!$omp critical
   call add_adjacencies(eqlist,eq-1,adjacencies,numadj,nentry)
!$omp end critical
end do
!$omp end parallel do

! Allocate column_index.
! Any changes to this list should be matched in the deallocate statement
! in destroy_linear_system

allocate(linear_system%column_index(nentry*syssize*syssize), stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation for matrix failed",procs=procs)
   return
endif

! Make a list of the objects the equations come from and the first
! index in column_index associated with each object

! numobj is larger than the actual number of objects because some
! edges and faces get counted more than once

numobj = grid%nvert + nelem_leaf*EDGES_PER_ELEMENT + nelem_leaf
if (global_element_kind == TETRAHEDRAL_ELEMENT) then
   numobj = numobj + nelem_leaf*FACES_PER_ELEMENT
endif
allocate(objtype(numobj),objid(numobj),first_indx(numobj),stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation for matrix failed",procs=procs)
   return
endif

! first the equations from vertices, level by level

iobj = 1
indx = 1

! Mark vertices in floating regions, except the head of the list, as already
! visited so they are not included.

visited_vert = .false.
if (associated(linear_system%floating_head)) then
   do region=1,size(linear_system%floating_head)
      vert = linear_system%floating_head(region)
      vert = linear_system%floating_next(vert)
      do while (vert /= END_OF_LIST)
         visited_vert(vert) = .true.
         vert = linear_system%floating_next(vert)
      end do
   end do
endif

do lev=1,grid%nlev
   elem = grid%head_level_elem(lev)
   do while (elem /= END_OF_LIST)
      do ivert=1,VERTICES_PER_ELEMENT
         vert = grid%element(elem)%vertex(ivert)
         if (visited_vert(vert)) cycle
         visited_vert(vert) = .true.
         objtype(iobj) = VERTEX_ID
         objid(iobj) = vert
         first_indx(iobj) = indx
         call grid_to_eq(grid,linear_system,VERTEX_ID,1,1, &
                         grid%vertex(vert)%gid,eq1)
         do i=1,syssize
            if (is_periodic_vert_slave(vert,grid,i)) cycle
            indx = indx + syssize*numadj(eq1)
         end do
         iobj = iobj + 1
      end do
      elem = grid%element(elem)%next
   end do
end do

! then the edge bases

visited_edge = .false.
do l=1,nelem_leaf
   elem = leaf_element(l)
   do j=1,EDGES_PER_ELEMENT
      edge = grid%element(elem)%edge(j)
      if (visited_edge(edge)) cycle
      visited_edge(edge) = .true.
      if (grid%edge(edge)%degree < 2) cycle
      objtype(iobj) = EDGE_ID
      objid(iobj) = edge
      first_indx(iobj) = indx
      call grid_to_eq(grid,linear_system,EDGE_ID,1,1, &
                      grid%edge(edge)%gid,eq1)
      do k=1,grid%edge(edge)%degree - 1
         do i=1,syssize
            if (is_periodic_edge_slave(edge,grid,i)) cycle
            indx = indx + syssize*numadj(eq1)
         end do
      end do
      iobj = iobj + 1
   end do
end do

! then the face bases

if (global_element_kind == TETRAHEDRAL_ELEMENT) then
   visited_face = .false.
   do l=1,nelem_leaf
      elem = leaf_element(l)
      do j=1,FACES_PER_ELEMENT
         face = grid%element(elem)%face(j)
         if (visited_face(face)) cycle
         visited_face(face) = .true.
         if (grid%face(face)%degree < 3) cycle
         objtype(iobj) = FACE_ID
         objid(iobj) = face
         first_indx(iobj) = indx
         call grid_to_eq(grid,linear_system,FACE_ID,1,1, &
                         grid%face(face)%gid,eq1)
         do k=1,face_dof(grid%face(face)%degree)
            do i=1,syssize
               if (is_periodic_face_slave(face,grid,i)) cycle
               indx = indx + syssize*numadj(eq1)
            end do
         end do
         iobj = iobj + 1
      end do
   end do
endif

! finally the bubble bases

do l=1,nelem_leaf
   elem = leaf_element(l)
   if (grid%element(elem)%degree < 3) cycle
   objtype(iobj) = ELEMENT_ID
   objid(iobj) = elem
   first_indx(iobj) = indx
   call grid_to_eq(grid,linear_system,ELEMENT_ID,1,1, &
                   grid%element(elem)%gid,eq1)
   do k=1,element_dof(grid%element(elem)%degree)
      do i=1,syssize
         indx = indx + syssize*numadj(eq1)
      end do
   end do
   iobj = iobj + 1
end do

numobj = iobj - 1

! Build column_index, begin_row, end_row_linear, end_row_edge, end_row_face
! and end_row.

!$omp parallel do schedule(dynamic) &
!$omp  default(shared) &
!$omp  private(iobj,indx,numdof,k,eq1,eq,i,m,objtype2,brank,srank)

do iobj = 1,numobj

   indx = first_indx(iobj)
   select case(objtype(iobj))
   case (VERTEX_ID)
      numdof = 1
   case (EDGE_ID)
      numdof = grid%edge(objid(iobj))%degree - 1
   case (FACE_ID)
      numdof = face_dof(grid%face(objid(iobj))%degree)
   case (ELEMENT_ID)
      numdof = element_dof(grid%element(objid(iobj))%degree)
   end select

   do k=1,numdof

      select case(objtype(iobj))
      case (VERTEX_ID)
         call grid_to_eq(grid,linear_system,VERTEX_ID,1,1, &
                         grid%vertex(objid(iobj))%gid,eq1)
      case (EDGE_ID)
         call grid_to_eq(grid,linear_system,EDGE_ID,k,1, &
                         grid%edge(objid(iobj))%gid,eq1)
      case (FACE_ID)
         call grid_to_eq(grid,linear_system,FACE_ID,k,1, &
                         grid%face(objid(iobj))%gid,eq1)
      case (ELEMENT_ID)
         call grid_to_eq(grid,linear_system,ELEMENT_ID,k,1, &
                         grid%element(objid(iobj))%gid,eq1)
      end select

      eq = eq1 - 1
      do i=1,syssize
         if (objtype(iobj) == VERTEX_ID) then
            if (is_periodic_vert_slave(objid(iobj),grid,i)) cycle
         endif
         if (objtype(iobj) == EDGE_ID) then
            if (is_periodic_edge_slave(objid(iobj),grid,i)) cycle
         endif
         if (objtype(iobj) == FACE_ID) then
            if (is_periodic_face_slave(objid(iobj),grid,i)) cycle
         endif
         eq = eq + 1
         linear_system%begin_row(eq) = indx
         do m=1,numadj(eq1)
            linear_system%column_index(indx:indx+syssize-1) = &
                                       (/(adjacencies(m,eq1)+j,j=0,syssize-1)/)
! for the diagonal block (which comes first in the row storage), swap so that
! the diagonal entry is the first entry of the row
            if (indx < linear_system%begin_row(eq) + syssize) then
              linear_system%column_index(indx) = adjacencies(m,eq1)+i-1
              linear_system%column_index(indx+i-1) = adjacencies(m,eq1)
            endif
! if this block is from a linear basis, set end_row_linear so it is the last
! linear basis column in the end
            call eq_to_grid(linear_system,linear_system%gid(adjacencies(m,eq1)), &
                            objtype2,brank,srank)
            if (objtype2 == VERTEX_ID) then
               linear_system%end_row_linear(eq) = indx + syssize - 1
            endif
            if (objtype2 == VERTEX_ID .or. objtype2 == EDGE_ID) then
               linear_system%end_row_edge(eq) = indx + syssize - 1
            endif
            if (global_element_kind == TETRAHEDRAL_ELEMENT) then
               if (objtype2 == VERTEX_ID .or. objtype2 == EDGE_ID .or. &
                   objtype2 == FACE_ID) then
                  linear_system%end_row_face(eq) = indx + syssize - 1
               endif
            endif
            indx = indx + syssize
         end do
         linear_system%end_row(eq) = indx - 1
      end do
   end do
end do
!$omp end parallel do

linear_system%begin_row(neq+1) = nentry*syssize*syssize+1
linear_system%end_row(neq+1) = nentry*syssize*syssize+1
linear_system%end_row_linear(neq+1) = nentry*syssize*syssize+1
linear_system%end_row_edge(neq+1) = nentry*syssize*syssize+1
if (global_element_kind == TETRAHEDRAL_ELEMENT) then
   linear_system%end_row_face(neq+1) = nentry*syssize*syssize+1
endif

! free up the memory used by adjacencies

deallocate(adjacencies)
deallocate(objtype,objid,first_indx)

end subroutine old_create_column_index

!        -----------------------
function find_max_entries_in_row(grid,leaf_element,nelem_leaf)
!        -----------------------

!----------------------------------------------------
! This routine determines the maximum number of entries in any row when
! system_size is 1 (otherwise multiply by system_size).  In 2D it is exactly
! the maximum, in 3D it is an upper bound.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: leaf_element(:), nelem_leaf
integer :: find_max_entries_in_row
!----------------------------------------------------
! Local variables:

integer :: num_adjacent(grid%biggest_vert)
integer :: elem, i, j, deg, multiplier, ielem
integer :: v(VERTICES_PER_ELEMENT), neigh(NEIGHBORS_PER_ELEMENT)
!----------------------------------------------------
! Begin executable code

! The maximum will occur at a row corresponding to a vertex, so add up the
! number of overlapping bases for each vertex basis.  In 2D pass through the
! elements and for each vertex of the element add 1) twice the number of bubble
! bases, 2) the number of edge bases on adjacent edges, or twice the number if
! the edge is on the boundary, 3) twice the number of edge bases on the opposite
! edge, and 4) 1 for each adjacent vertex, or 2 if the vertex is on the
! boundary.  Then divide by 2.  And add 1 for the diagonal.
! The business with 2 takes care of duplications over adjacent elements.
! In 3D the trick with 2 doesn't work because an edge can be shared by more
! than two elements, so compute an upper bound by passing through the elements
! and for each vertex of the element add the number of bases in the element.

select case (global_element_kind)
case (TRIANGULAR_ELEMENT)
   multiplier = 2
case (TETRAHEDRAL_ELEMENT)
   multiplier = 1
end select

num_adjacent = 0

! for each leaf element

!$omp parallel do &
!$omp  default(shared) &
!$omp  private(ielem,elem,v,neigh,i,deg,j) &
!$omp  reduction(+ : num_adjacent)

do ielem=1,nelem_leaf
   elem = leaf_element(ielem)
   v = grid%element(elem)%vertex
   if (global_element_kind == TRIANGULAR_ELEMENT) then
      neigh = get_neighbors(elem,grid)
   endif
! for each vertex of this element
   do i=1,VERTICES_PER_ELEMENT
      do while(is_periodic_vert_slave(v(i),grid))
         v(i) = grid%vertex(v(i))%next
      end do
      deg = grid%element(elem)%degree
! bubble bases
      num_adjacent(v(i)) = num_adjacent(v(i)) + multiplier*element_dof(deg)
! face bases
      do j=1,FACES_PER_ELEMENT
         deg = grid%face(grid%element(elem)%face(j))%degree
         num_adjacent(v(i)) = num_adjacent(v(i)) + multiplier*face_dof(deg)
      end do
! edge bases
      do j=1,EDGES_PER_ELEMENT
         deg = grid%edge(grid%element(elem)%edge(j))%degree
         if (j==i .or. global_element_kind == TETRAHEDRAL_ELEMENT) then
! opposite edge in 2D, all edges in 3D
            num_adjacent(v(i)) = num_adjacent(v(i)) + multiplier*(deg-1)
         else
! adjacent edges in 2D
            if (neigh(j)==BOUNDARY) then
               num_adjacent(v(i)) = num_adjacent(v(i)) + multiplier*(deg-1)
            else
               num_adjacent(v(i)) = num_adjacent(v(i)) + deg-1
            endif
! adjacent vertices in 2D
            if (neigh(6-(i+j))==BOUNDARY) then
               num_adjacent(v(i)) = num_adjacent(v(i)) + 2
            else
               num_adjacent(v(i)) = num_adjacent(v(i)) + 1
            endif
         endif
      end do
! vertices in 3D
      if (global_element_kind == TETRAHEDRAL_ELEMENT) then
         num_adjacent(v(i)) = num_adjacent(v(i)) + 3
      endif
   end do
end do
!$omp end parallel do

find_max_entries_in_row = maxval(num_adjacent)/multiplier + 1

end function find_max_entries_in_row

!          ---------------
subroutine add_adjacencies(list,n,adjacencies,numadj,nentry)
!          ---------------

!----------------------------------------------------
! This routine takes a list of equations and makes sure all pairs are on the
! lists of adjacencies
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: list(:), n
integer, intent(inout) :: adjacencies(:,:), numadj(:)
integer, intent(inout) :: nentry

!----------------------------------------------------
! Local variables:

integer :: i, j, k, list_eq, adj_eq, entry, lo, hi
!----------------------------------------------------
! Begin executable code

! For each ordered pair of list entries

do i=1,n
   list_eq = list(i)

! The first time a row is seen, place the diagonal as the first entry

   if (numadj(list_eq) == 0) then
      adjacencies(1,list_eq) = list_eq
      numadj(list_eq) = 1
      nentry = nentry + 1
   endif

   do j=1,n
      adj_eq = list(j)

      if (list_eq == adj_eq) cycle

! use a binary search to insert adj_eq into the adjacencies of list_eq in
! increasing order (except list_eq is first)

! if it is larger than the last entry, put it at the end

      if (numadj(list_eq) == 1 .or. &
          adj_eq > adjacencies(numadj(list_eq),list_eq)) then
         numadj(list_eq) = numadj(list_eq) + 1
         adjacencies(numadj(list_eq),list_eq) = adj_eq
         nentry = nentry + 1
         cycle
      endif

! set the range to search and start in the middle

      lo = 2
      hi = numadj(list_eq)
      entry = (lo+hi)/2

      do

! if equal, already on list

         if (adj_eq == adjacencies(entry,list_eq)) exit

! if it is less than the current entry and this entry or the previous entry is
! the low end of the range, then insert it before the current entry or before
! the low end of the range

         if (adj_eq < adjacencies(entry,list_eq)) then

            if (entry == lo .or. entry-1 == lo) then
               if (adj_eq == adjacencies(lo,list_eq)) exit
               if (adj_eq < adjacencies(lo,list_eq)) entry = lo
               do k=numadj(list_eq),entry,-1
                  adjacencies(k+1,list_eq) = adjacencies(k,list_eq)
               end do
               adjacencies(entry,list_eq) = adj_eq
               numadj(list_eq) = numadj(list_eq) + 1
               nentry = nentry + 1
               exit

! if it is less than the current entry and we're not at the low end, reset
! the high end and go to the middle of the new range

            else
               hi = entry
               entry = (lo+hi)/2
            endif

! if it is greater than the current entry and this entry or the next entry is
! the high end of the range, then insert it after the current entry or after
! the high end of the range

         else

            if (entry == hi .or. entry+1 == hi) then
               if (adj_eq == adjacencies(hi,list_eq)) exit
               if (adj_eq > adjacencies(hi,list_eq)) entry = hi
               entry = entry + 1
               do k=numadj(list_eq),entry,-1
                  adjacencies(k+1,list_eq) = adjacencies(k,list_eq)
               end do
               adjacencies(entry,list_eq) = adj_eq
               numadj(list_eq) = numadj(list_eq) + 1
               nentry = nentry + 1
               exit

! if it is greater than the current entry and we're not at the high end, reset
! the low end and go to the middle of the new range

            else
               lo = entry
               entry = (lo+hi)/2
            endif

         endif
      end do
   end do
end do

end subroutine add_adjacencies

!          -------------------
subroutine create_column_index(linear_system,grid,procs,maxeq_per_elem, &
                               leaf_element,nelem_leaf)
!          -------------------

!----------------------------------------------------
! This routine creates column_index and related arrays.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linear_system
type(grid_type), intent(inout) :: grid
type(proc_info), intent(in) :: procs
integer, intent(in) :: maxeq_per_elem, leaf_element(:), nelem_leaf
!----------------------------------------------------
! Local variables:

integer :: i, j, ss, lasteq_vert, lasteq_edge, allocsize, nentry, ielem, elem, &
           neq, astat, jerr, eq, last_assigned, end_row, ind, &
           sr, sc, this_eq, ncol, entry, last_entry, max_cols_per_eq
integer, allocatable :: lotsa_cols(:), next(:), elem_eq(:), iperm(:), &
                        column_list(:), cols(:), begin_row(:), &
                        end_row_linear(:), end_row_edge(:), end_row_bubble(:), &
                        eq_last_entry(:), cols_per_eq(:)
!----------------------------------------------------
! Begin executable code

! the old version is better, especially in memory, for 3D

if (global_element_kind == TETRAHEDRAL_ELEMENT) then
   call old_create_column_index(linear_system,grid,procs,maxeq_per_elem, &
                                leaf_element,nelem_leaf)
   return
endif

! convenience variables

ss = linear_system%system_size
lasteq_vert = linear_system%neq_vert
lasteq_edge = lasteq_vert + linear_system%neq_edge

! initializations

linear_system%end_row_linear = 0
linear_system%end_row_edge = 0

! Create an array, lotsa_cols, that will contain the columns for each row, and
! an array, next, that, for each entry in lotsa_cols, points to the next entry
! of lotsa_cols for the same row.  The first neq entries of lotsa_cols contain
! the diagonal of the matrix to get each linked list started.  The lists will
! contain the columns found from each element, so there will be duplicates.
! For systems of equations, just do the first one and expand it when copying
! to column_index in the end.
! Also create an array of size maxeq_per_elem to list the equation lids in an
! element.  And create arrays for the beginning and end of rows.

allocsize = nelem_leaf*maxeq_per_elem**2
allocate(lotsa_cols(allocsize),next(allocsize),elem_eq(maxeq_per_elem), &
         eq_last_entry(linear_system%neq/ss), &
         cols_per_eq(linear_system%neq/ss), &
         begin_row(linear_system%neq/ss+1), &
         end_row_linear(linear_system%neq/ss+1), &
         end_row_edge(linear_system%neq/ss+1), &
         end_row_bubble(linear_system%neq/ss+1), &
         stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("create_column_index: memory allocation failed",procs=procs)
   stop
endif
lotsa_cols(1:linear_system%neq/ss) = (/ (ss*(i-1)+1,i=1,linear_system%neq/ss) /)
eq_last_entry = (/ (i,i=1,linear_system%neq/ss) /)
next = -1
cols_per_eq = 1
end_row_linear = 0
end_row_edge = 0

! Go through the elements adding the equation contributions of each element.

nentry = linear_system%neq/ss
do ielem=1,nelem_leaf
   elem = leaf_element(ielem)

! Get the equation lids of this element

   call make_elem_eq_list(grid,linear_system,elem,elem_eq,neq)

! For each row

   do i=1,neq

! Find the end of the linked list for this row

      last_entry = eq_last_entry((elem_eq(i)-1)/ss+1)

! Add each of the columns for this row, except for self because self is
! already at the beginning of the list

      do j=1,neq
         if (elem_eq(j) == elem_eq(i)) cycle
         nentry = nentry + 1
         next(last_entry) = nentry
         lotsa_cols(nentry) = elem_eq(j)
         last_entry = nentry
         cols_per_eq((elem_eq(i)-1)/ss+1) = cols_per_eq((elem_eq(i)-1)/ss+1)+1
      end do
      eq_last_entry((elem_eq(i)-1)/ss+1) = last_entry
   end do
end do

max_cols_per_eq = maxval(cols_per_eq)

! Done with these allocatables

deallocate(elem_eq,eq_last_entry,cols_per_eq)

! Allocate column_list to contain the list of columns for a row as extracted
! from lotsa_cols, a permutation vector for sort(column_list), and cols which
! will contain the sorted columns with duplicates removed.

allocate(column_list(max_cols_per_eq),iperm(max_cols_per_eq),cols(nentry), &
         stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("create_column_index: memory allocation failed",procs=procs)
   stop
endif

! For each equation, extract the linked list of columns, sort them, and
! eliminate duplicates.  Also set begin_row and the end_row's.

end_row = 0
do eq=1,linear_system%neq/ss
   this_eq = ss*(eq-1)+1

! Extract the list of columns for this row

   ncol = 0
   entry = eq
   do while (next(entry) /= -1)
      ncol = ncol + 1
      column_list(ncol) = lotsa_cols(entry)
      entry = next(entry)
   end do
   ncol = ncol + 1
   column_list(ncol) = lotsa_cols(entry)

! Sort column_list

   call sort(column_list,ncol,iperm,2,jerr)
   if (jerr /= 0) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("create_column_index: sort returned error code", &
                 intlist=(/jerr/),procs=procs)
      stop
   endif

! Set the beginning of this row

   begin_row(eq) = end_row + 1

! Eliminate duplicates by copying the first of each column index to the
! location where it will be in column_index.  In the process, set the end_row's.
! Also, when the diagonal entry is encountered, move it to the first position
! and shift those already assigned up one.

   cols(begin_row(eq)) = column_list(1)
   last_assigned = column_list(1)
   end_row = begin_row(eq)
   do i=1,ncol

      if (column_list(i) /= last_assigned) then

         end_row = end_row + 1

         if (column_list(i) == this_eq) then
            cols(end_row) = cols(begin_row(eq))
            cols(begin_row(eq)+1:end_row) = &
               cols(begin_row(eq):end_row-1)
            cols(begin_row(eq)) = column_list(i)
         else
            cols(end_row) = column_list(i)
         endif
         last_assigned = column_list(i)

         if (column_list(i)>lasteq_vert .and. end_row_linear(eq)==0) then
            if (eq > lasteq_vert/ss) then
               end_row_linear(eq) = end_row
            else
               end_row_linear(eq) = end_row - 1
            endif
         endif

         if (column_list(i)>lasteq_edge .and. end_row_edge(eq)==0) then
            if (eq > lasteq_edge/ss) then
               end_row_edge(eq) = end_row
            else
               end_row_edge(eq) = end_row - 1
            endif
         endif

      endif

   end do
   if (end_row_linear(eq)==0) then
      end_row_linear(eq) = end_row
   endif
   if (end_row_edge(eq)==0) then
      end_row_edge(eq) = end_row
   endif
   end_row_bubble(eq) = end_row

end do ! next equation

! Set an extra begin_row and end_rows

begin_row(linear_system%neq/ss+1) = end_row+1
end_row_linear(linear_system%neq/ss+1) = end_row+1
end_row_edge(linear_system%neq/ss+1) = end_row+1
end_row_bubble(linear_system%neq/ss+1) = end_row+1

! Done with these allocatables

deallocate(lotsa_cols, next, iperm, column_list)

! Allocate column_index

allocate(linear_system%column_index(end_row*ss*ss),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("create_column_index: memory allocation failed",procs=procs)
   stop
endif

! Copy the column indices to column_index

! For a scalar equation, just copy it

if (ss == 1) then
   linear_system%column_index(1:end_row) = cols(1:end_row)
   linear_system%begin_row = begin_row
   linear_system%end_row_linear = end_row_linear
   linear_system%end_row_edge = end_row_edge
   linear_system%end_row_bubble = end_row_bubble

! For systems of equations, each entry gets expanded to an ss X ss block

else

! ind is the current index in linear_system%column_index

   ind = 0

! for each block of equations ...

   do eq=1,linear_system%neq/ss

! for each of the ss rows of this block ...

      do sr=1,ss

! this row starts at the next index

         linear_system%begin_row(ss*(eq-1)+sr) = ind+1

! for each of the unexpanded column indices

         do i=begin_row(eq),end_row_bubble(eq)

! for each of the ss columns of this block

            do sc=1,ss

! set the column index

               ind = ind + 1
               linear_system%column_index(ind) = cols(i) + sc-1

! check for end_row's

               if (end_row_linear(eq) == i .and. sc==ss) then
                  linear_system%end_row_linear(ss*(eq-1)+sr) = ind
               endif
               if (end_row_edge(eq) == i .and. sc==ss) then
                  linear_system%end_row_edge(ss*(eq-1)+sr) = ind
               endif
            end do

! swap the diagonal to the beginning of the row

            if (i == begin_row(eq) .and. sr /= 1) then
               linear_system%column_index(ind-ss+2:ind-ss+sr) = &
                  linear_system%column_index(ind-ss+1:ind-ss+sr-1)
               linear_system%column_index(ind-ss+1) = cols(begin_row(eq)) + sr-1
            endif
         end do

! set the final end of row for this row

         linear_system%end_row_bubble(ss*(eq-1)+sr) = ind
      end do
   end do

! Set an extra begin_row and end_rows

   linear_system%begin_row(linear_system%neq+1) = ind+1
   linear_system%end_row_linear(linear_system%neq+1) = ind+1
   linear_system%end_row_edge(linear_system%neq+1) = ind+1
   linear_system%end_row_bubble(linear_system%neq+1) = ind+1

endif

! Free work space

deallocate(cols,begin_row,end_row_linear,end_row_edge,end_row_bubble)

end subroutine create_column_index

!          -----------------
subroutine make_elem_eq_list(grid,linear_system,elem,elem_eq,neq)
!          -----------------

!----------------------------------------------------
! This routine returns a list of the equation lids associated with element
! elem.  For systems of equations, it only returns the first one.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
type(linsys_type), intent(in) :: linear_system
integer, intent(in) :: elem
integer, intent(out) :: elem_eq(:), neq
!----------------------------------------------------
! Local variables:

integer :: i, brank
!----------------------------------------------------
! Begin executable code

neq = 0
do i=1,VERTICES_PER_ELEMENT
   neq = neq + 1
   call grid_to_eq(grid,linear_system,VERTEX_ID,1,1, &
                   grid%vertex(grid%element(elem)%vertex(i))%gid,elem_eq(neq))
end do

do i=1,EDGES_PER_ELEMENT
   do brank=1,grid%edge(grid%element(elem)%edge(i))%degree-1
      neq = neq + 1
      call grid_to_eq(grid,linear_system,EDGE_ID,brank,1, &
                      grid%edge(grid%element(elem)%edge(i))%gid,elem_eq(neq))
   end do
end do

do i=1,FACES_PER_ELEMENT
   do brank=1,face_dof(grid%face(grid%element(elem)%face(i))%degree)
      neq = neq + 1
      call grid_to_eq(grid,linear_system,FACE_ID,brank,1, &
                      grid%face(grid%element(elem)%face(i))%gid,elem_eq(neq))
   end do
end do

do i=1,element_dof(grid%element(elem)%degree)
   neq = neq + 1
   call grid_to_eq(grid,linear_system,ELEMENT_ID,i,1, &
                   grid%element(elem)%gid,elem_eq(neq))
end do

end subroutine make_elem_eq_list

!          ---------------
subroutine allocate_matrix(linear_system,solver_cntl,io_cntl,procs)
!          ---------------

!----------------------------------------------------
! This routine allocates the matrix value arrays and initializes them and
! the right hand side arrays to 0.0
! Any changes to this list should be matched in the deallocate statement
! in destroy_linear_system
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linear_system
type(solver_options), intent(in) :: solver_cntl
type(io_options), intent(in) :: io_cntl
type(proc_info), intent(in) :: procs
!----------------------------------------------------
! Local variables:

integer :: siz, allocstat, neq
!----------------------------------------------------
! Begin executable code

siz = size(linear_system%column_index)
neq = linear_system%neq

! Allocate stiffness matrix.

allocate(linear_system%stiffness(siz), stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation for matrix failed",procs=procs)
   return
endif
linear_system%stiffness  = 0.0_my_real

! Allocate mass matrix and if this is an eigenvalue problem.

if (solver_cntl%eq_type == EIGENVALUE) then
   allocate(linear_system%mass(siz), stat=allocstat)
   if (allocstat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation for matrix failed",procs=procs)
      return
   endif
   linear_system%mass = 0.0_my_real
else
   nullify(linear_system%mass)
endif

! If this is a boundary value problem, the matrix to be statically condensed
! is the stiffness matrix  But if the residual is to be computed
! (print_error_when is FREQUENTLY or TOO_MUCH) we need to keep the stiffness
! matrix and use different space for condensed.
! If this is an eigenvalue problem, static condensation is not performed.

if (solver_cntl%eq_type == EIGENVALUE) then
   nullify(linear_system%condensed)
   nullify(linear_system%rhs_cond)
else
   if (io_cntl%print_error_when == FREQUENTLY .or. &
       io_cntl%print_error_when == TOO_MUCH) then
      allocate(linear_system%condensed(siz),linear_system%rhs_cond(neq), &
               stat=allocstat)
      if (allocstat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation for matrix failed",procs=procs)
         return
      endif
   else
      linear_system%condensed => linear_system%stiffness
      linear_system%rhs_cond => linear_system%rhs_nocond
   endif
endif

! Initialize rhs related arrays to 0

linear_system%rhs        = 0.0_my_real
linear_system%r_mine     = 0.0_my_real
linear_system%r_others   = 0.0_my_real
linear_system%need_r_others = .false.

end subroutine allocate_matrix

!          ------------------
subroutine compute_matrix_rhs (linear_system,grid,procs,solver_cntl, &
                               maxeq_per_elem,leaf_element,nelem_leaf, &
                               timeit,still_sequential)
!          ------------------

!----------------------------------------------------
! This routine computes the matrix and rhs values
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linear_system
type(grid_type), intent(in) :: grid
type(proc_info), intent(in) :: procs
type(solver_options), intent(in) :: solver_cntl
integer, intent(in) :: maxeq_per_elem,leaf_element(:),nelem_leaf
logical, intent(in) :: timeit, still_sequential
!----------------------------------------------------
! Local variables:

integer :: i, siz, elem, nleq, astat1, astat2, lid
integer, allocatable :: eqlist(:)
real(my_real), allocatable :: local_stiffness(:,:), local_mass(:,:), &
                              local_rhs(:)
!----------------------------------------------------
! Begin executable code

siz = maxeq_per_elem*linear_system%system_size

!$omp parallel &
!$omp  default(shared) &
!$omp  private(i, elem, eqlist, local_stiffness, local_mass, local_rhs, &
!$omp          nleq, astat1, astat2)

! space for equation list and elemental matrices

allocate(eqlist(maxeq_per_elem),local_rhs(siz),stat=astat1)
if (.not. linear_system%linsys_exists) then
   allocate(local_stiffness(siz,siz),local_mass(siz,siz),stat=astat2)
else
   astat2 = 0
endif
if (astat1 /= 0 .or. astat2 /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in compute_matrix_rhs")
   stop
endif

! pass through the elements

! OpenMP parallel section.  To get the same results as a sequential program,
! we need the ordered construct for assembling the global matrix, because
! floating point arithmetic is not commutative.  To keep this somewhat
! efficient we need a chunk size of 1 (so threads aren't waiting for the
! previous thread to complete many loops), and since the amount of work in
! computing the elemental matrix varies strongly with the degree of the element,
! we use a dynamic schedule (which has a default chunk size of 1).

!$omp do ordered, schedule(dynamic)
do i=1,nelem_leaf
   elem = leaf_element(i)

! compute the local matrix and right hand side

   call compute_local_matrix(grid,linear_system,solver_cntl,elem, &
                             local_stiffness,local_rhs,local_mass,eqlist,nleq)

! unless we are ignoring quadrature errors, remove the linear basis
! contributions if I don't own the element.  They will be gotten from the
! owner of the element in fix_quad_err.

   if (.not. solver_cntl%ignore_quad_err .and. &
      (.not. grid%element(elem)%iown)) then
      if (.not. linear_system%linsys_exists) then
         local_stiffness(1:VERTICES_PER_ELEMENT*linear_system%system_size, &
                         1:VERTICES_PER_ELEMENT*linear_system%system_size) =  &
                         0.0_my_real
         local_mass(1:VERTICES_PER_ELEMENT*linear_system%system_size, &
                    1:VERTICES_PER_ELEMENT*linear_system%system_size) = &
                    0.0_my_real
      endif
      local_rhs(1:VERTICES_PER_ELEMENT*linear_system%system_size) = 0.0_my_real
   endif

! assemble local values into global matrix etc.

!$omp ordered
   call assemble(linear_system,eqlist,nleq,local_stiffness,local_rhs, &
                 local_mass)
!$omp end ordered

end do
!$omp end do

deallocate(eqlist, local_rhs)
if (.not. linear_system%linsys_exists) then
   deallocate(local_stiffness, local_mass)
endif

!$omp end parallel

! fix the quadrature errors

if (.not. solver_cntl%ignore_quad_err) then
   call fix_quad_error(grid,procs,linear_system,still_sequential,timeit)
endif

! Adjust the rhs of rows for floating regions by the total charge

if (associated(grid%floating_tags)) then
   do i=1,size(grid%floating_tags)
      call grid_to_eq(grid,linear_system,VERTEX_ID,1,1, &
                      grid%vertex(linear_system%floating_head(i))%gid,lid)
      linear_system%rhs(lid) = linear_system%rhs(lid) + &
                               grid%floating_total_charge(i)*chargeUnit
   end do
endif

end subroutine compute_matrix_rhs

!          --------------------
subroutine compute_local_matrix(grid,linear_system,solver_cntl,elem,lmat, &
                                lrhs,lmassmat,eqlist,nleq)
!          --------------------

!----------------------------------------------------
! This routine computes the local stiffness and mass matrices and right side
! over element elem.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
type(linsys_type), intent(inout) :: linear_system
type(solver_options), intent(in) :: solver_cntl
integer, intent(in) :: elem
real(my_real), intent(out) :: lmat(:,:),lrhs(:),lmassmat(:,:)
integer, intent(out) :: eqlist(:), nleq

!----------------------------------------------------
! Local variables:

real(my_real) :: xvert(VERTICES_PER_ELEMENT), yvert(VERTICES_PER_ELEMENT), &
                 zvert(VERTICES_PER_ELEMENT)
integer :: i,j,k,deg
integer :: degree(EDGES_PER_ELEMENT+FACES_PER_ELEMENT+1), &
           bmark(EDGES_PER_ELEMENT+FACES_PER_ELEMENT), &
           edge_type(EDGES_PER_ELEMENT,linear_system%system_size), &
           face_type(FACES_PER_ELEMENT,linear_system%system_size)

!----------------------------------------------------
! Begin executable code

! determine local IDs of the equations

nleq = 1
do j=1,VERTICES_PER_ELEMENT
   call grid_to_eq(grid,linear_system,VERTEX_ID,1,1, &
                   grid%vertex(grid%element(elem)%vertex(j))%gid,eqlist(nleq))
   nleq = nleq + 1
end do
do j=1,EDGES_PER_ELEMENT
   do k = 1,grid%edge(grid%element(elem)%edge(j))%degree-1
      call grid_to_eq(grid,linear_system,EDGE_ID,k,1, &
                      grid%edge(grid%element(elem)%edge(j))%gid,eqlist(nleq))
      nleq = nleq + 1
   end do
end do
do j=1,FACES_PER_ELEMENT
   do k = 1,face_dof(grid%face(grid%element(elem)%face(j))%degree)
      call grid_to_eq(grid,linear_system,FACE_ID,k,1, &
                      grid%face(grid%element(elem)%face(j))%gid,eqlist(nleq))
      nleq = nleq + 1
   end do
end do
deg = grid%element(elem)%degree
do j=1,element_dof(deg)
   call grid_to_eq(grid,linear_system,ELEMENT_ID,j,1, &
                   grid%element(elem)%gid,eqlist(nleq))
   nleq = nleq + 1
end do
nleq = nleq - 1

! get vertex coordinates

do i=1,VERTICES_PER_ELEMENT
   xvert(i) = grid%vertex(grid%element(elem)%vertex(i))%coord%x
   yvert(i) = grid%vertex(grid%element(elem)%vertex(i))%coord%y
   zvert(i) = zcoord(grid%vertex(grid%element(elem)%vertex(i))%coord)
end do

! set degree, edge_type and bmark from the element

do i=1,EDGES_PER_ELEMENT
   degree(i) = grid%edge(grid%element(elem)%edge(i))%degree
   edge_type(i,:) = grid%edge_type(grid%element(elem)%edge(i),:)
   bmark(i) = grid%edge(grid%element(elem)%edge(i))%bmark
end do
do i=1,FACES_PER_ELEMENT
   degree(EDGES_PER_ELEMENT+i) = grid%face(grid%element(elem)%face(i))%degree
   face_type(i,:) = grid%face_type(grid%element(elem)%face(i),:)
   bmark(EDGES_PER_ELEMENT+i) = grid%face(grid%element(elem)%face(i))%bmark
end do
degree(EDGES_PER_ELEMENT+FACES_PER_ELEMENT+1) = grid%element(elem)%degree

if (associated(linear_system%mass)) then
   call elemental_matrix(grid, xvert, yvert, degree, edge_type, face_type, &
                         bmark, linear_system%system_size, &
                         solver_cntl%inc_quad_order,.false., elem, &
                         "p","a",lmat, lrhs, &
                         loc_bconds_s=my_bconds,lmassmat=lmassmat,zvert=zvert, &
                         solver_control=solver_cntl, &
                         just_rhs=linear_system%linsys_exists)
else
   call elemental_matrix(grid, xvert, yvert, degree, edge_type, face_type, &
                         bmark,linear_system%system_size, &
                         solver_cntl%inc_quad_order,.false., elem, &
                         "p","a",lmat, lrhs, &
                         loc_bconds_s=my_bconds,zvert=zvert, &
                         solver_control=solver_cntl, &
                         just_rhs=linear_system%linsys_exists)
endif

end subroutine compute_local_matrix

!          --------------------
subroutine elemental_matrix(grid, xvert, yvert, degree, edge_type, face_type, &
                            bmark, ss, inc_quad_order, resid_prob, elem, &
                            which_basis, which_set, lmat, lrhs, &
                            loc_bconds_a, loc_bconds_s, lmassmat, extra_rhs, &
                            zvert,solver_control,just_rhs,llapmat)
!          --------------------

!----------------------------------------------------
! This routine computes the local stiffness matrix and right hand side, and
! optionally mass matrix over the element with the given vertices and
! degrees, using the pde coefficient functions and given boundary conditions.
! extra_rhs is used for multiple right hand sides, currently supports only
! residual problems for multiple eigenvalues. Dimensions of extra_rhs are
! (system_size*nbasis,num_eval-1).  The first eigenvalue goes in lrhs and
! eigenvalues 2 to num_eval are in order in extra_rhs.
! which_basis is "n" for nodal, "h" for h-hierarchical, "p" for p-hierarchical
! which_set is "r" for red bases, "b" for black bases and "a" for all.
! llapmat returns the local matrix for the Laplacian; it is used for H1 norms.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
real(my_real), intent(in) :: xvert(:), yvert(:)
integer, intent(in) :: degree(:), edge_type(:,:), face_type(:,:), bmark(:)
integer, intent(in) :: ss ! system_size
integer, intent(in) :: inc_quad_order, elem
logical, intent(in) :: resid_prob
character(len=1), intent(in) :: which_basis, which_set
real(my_real), intent(out) :: lmat(:,:),lrhs(:)
interface
   subroutine loc_bconds_a(x,y,z,bmark,itype,c,rs,extra_rs)
   use global
   real(my_real), intent(in) :: x(:),y(:),z(:)
   integer, intent(in) :: bmark
   integer, intent(out) :: itype(:)
   real(my_real), intent(out) :: c(:,:,:),rs(:,:)
   real(my_real), intent(out), optional :: extra_rs(:,:,:)
   end subroutine loc_bconds_a
   subroutine loc_bconds_s(x,y,z,bmark,itype,c,rs)
   use global
   real(my_real), intent(in) :: x,y,z
   integer, intent(in) :: bmark
   integer, intent(out) :: itype(:)
   real(my_real), intent(out) :: c(:,:),rs(:)
   end subroutine loc_bconds_s
end interface
optional :: loc_bconds_a, loc_bconds_s
real(my_real), intent(out), optional :: lmassmat(:,:), extra_rhs(:,:)
real(my_real), intent(in), optional :: zvert(:)
type(solver_options), intent(in), optional :: solver_control
logical, intent(in), optional :: just_rhs
real(my_real), intent(out), optional :: llapmat(:,:)

!----------------------------------------------------
! Local variables:

integer :: bctype(ss)
real(my_real) :: cxx(ss,ss), cyy(ss,ss), czz(ss,ss), cxy(ss,ss), cxz(ss,ss), &
                 cyz(ss,ss), cx(ss,ss), cy(ss,ss), cz(ss,ss), c(ss,ss), rs(ss)
real(my_real), allocatable, save :: basis(:,:),basisx(:,:),basisy(:,:), &
                                    basisz(:,:), basis_bound(:,:), &
                                    u(:,:,:),ux(:,:,:),uy(:,:,:),uz(:,:,:), &
                                    bcrs(:,:),bcc(:,:,:),extra_bcrs(:,:,:)
!$omp threadprivate(basis,basisx,basisy,basisz,basis_bound,u,ux,uy,uz, &
!$omp               bcrs,bcc,extra_bcrs)
real(my_real) :: xelem(3),yelem(3),zelem(3),term
integer :: i,j,k,l,qp,side,nquad_pts,jerr,astat,nbasis,qdegree,edim, &
           max_quad_order_elem,ilimit,astat1,astat2,add2side
logical :: doit, include_cross, include_first, use_laplacian_table, &
           not_just_rhs, use_cached_basis
logical, save :: warned=.false.
real(my_real), pointer :: xquad(:), yquad(:), zquad(:), quad_weight(:)
!----------------------------------------------------
! Begin executable code

! whether or not to include cross and first order derivative terms, and
! whether or not to use the table of integrals for the Laplacian operator.

if (present(solver_control)) then
   include_cross = solver_control%pde_has_cross_derivative
   include_first = solver_control%pde_has_first_order_terms
   use_laplacian_table = solver_control%laplacian_operator .and. &
                         solver_control%isosceles_right_triangles .and. &
                         which_set == "a" .and. which_basis == "p" .and. &
                         ss == 1 .and. .not. present(lmassmat)
else
   include_cross = .true.
   include_first = .true.
   use_laplacian_table = .false.
endif

! is this just an rhs update?

if (present(just_rhs)) then
   not_just_rhs = .not. just_rhs
else
   not_just_rhs = .true.
endif

! we can cache the bases at quadrature points and reuse them if
! 1) it is the p-hierarchical basis, and
! 2) we are doing both black and red bases

use_cached_basis = which_basis == "p" .and. which_set == "a"

! zvert must be present for 3D

if (global_element_kind == TETRAHEDRAL_ELEMENT .and. .not. present(zvert)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("zvert must be present in elemental matrix for 3D")
   stop
endif

! nodal and h-hierarchical bases require the same degree throughout

if (which_basis == "n" .or. which_basis == "h") then
   if (any(degree /= degree(1))) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("degree must be the same for edges and elements for nodal or h-hierarchical basis", &
                 intlist = degree)
      stop
   endif
endif

! number of basis functions over this element

select case (which_set)

case("a")
   nbasis = VERTICES_PER_ELEMENT
   do i=1,EDGES_PER_ELEMENT 
      nbasis = nbasis + max(0,degree(i)-1)
   end do
   do i=1,FACES_PER_ELEMENT
      nbasis = nbasis + face_dof(degree(EDGES_PER_ELEMENT+i))
   end do
   nbasis = nbasis + element_dof(degree(EDGES_PER_ELEMENT+FACES_PER_ELEMENT+1))

case("b")
   if (global_element_kind == TETRAHEDRAL_ELEMENT) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("elemental_matrix: doesn't support only black bases in 3D")
      stop
   endif
   if (which_basis == "p") then
      nbasis = VERTICES_PER_ELEMENT
      do i=1,EDGES_PER_ELEMENT
         nbasis = nbasis + max(0,degree(i)-2)
      end do
      if (degree(EDGES_PER_ELEMENT+FACES_PER_ELEMENT+1) > 3) then
         nbasis = nbasis + ((degree(EDGES_PER_ELEMENT+FACES_PER_ELEMENT+1)-3) * &
                            (degree(EDGES_PER_ELEMENT+FACES_PER_ELEMENT+1)-2))/2
      endif
   else
      if (2*(degree(1)/2) == degree(1)) then
         nbasis = ((degree(1)+2)*(degree(1)+2))/4
      else
         nbasis = ((degree(1)+1)*(degree(1)+3))/4
      endif
   endif

case("r")
   select case(global_element_kind)
   case (TRIANGULAR_ELEMENT)
      if (which_basis == "p") then
         nbasis = 0
         do i=1,EDGES_PER_ELEMENT
            if (degree(i) >= 2) nbasis = nbasis + 1
         end do
         if (degree(EDGES_PER_ELEMENT+FACES_PER_ELEMENT+1) >= 3) then
            nbasis = nbasis + degree(EDGES_PER_ELEMENT+FACES_PER_ELEMENT+1)-2
         endif
      else
         if (2*(degree(1)/2) == degree(1)) then
            nbasis = (degree(1)*(degree(1)+2))/4
         else
            nbasis = ((degree(1)+1)*(degree(1)+1))/4
         endif
      endif
   case (TETRAHEDRAL_ELEMENT)
      if (which_basis == "p") then
         nbasis = 0
         do i=1,EDGES_PER_ELEMENT
            if (degree(i) >= 2) nbasis = nbasis + 1
         end do
         do i=1,FACES_PER_ELEMENT
            if (degree(EDGES_PER_ELEMENT+i) >= 3) then
               nbasis = nbasis + degree(EDGES_PER_ELEMENT+i) - 2
            endif
         end do
         if (degree(EDGES_PER_ELEMENT+FACES_PER_ELEMENT+1) >= 4) then
            nbasis=nbasis+((degree(EDGES_PER_ELEMENT+FACES_PER_ELEMENT+1)-2)* &
                           (degree(EDGES_PER_ELEMENT+FACES_PER_ELEMENT+1)-3))/2
         endif
      else
         ierr = PHAML_INTERNAL_ERROR
         call fatal("elemental_matrix: only supports p-hierarchical basis in 3D")
         stop
      endif
   end select

end select

! determine the degree of the quadrature rule that is exact for
! polynomials of the same degree as the approximating space.

select case (global_element_kind)
case (TRIANGULAR_ELEMENT)
   max_quad_order_elem = MAX_QUAD_ORDER_TRI
case (TETRAHEDRAL_ELEMENT)
   max_quad_order_elem = MAX_QUAD_ORDER_TET
end select
qdegree = 2*maxval(degree)
if (qdegree > max_quad_order_elem) then
   if (.not. warned) then
      call warning("Element degree requires quadrature rule larger than available.", &
                   "Answers may not be as accurate as expected.")
      warned = .true.
   endif
   qdegree = max_quad_order_elem
endif
qdegree = qdegree + inc_quad_order
if (qdegree > max_quad_order_elem) qdegree = max_quad_order_elem
if (qdegree < 1) qdegree = 1

! Special code for SEM simulation right side update

if (present(just_rhs)) then
   if (sem_simulation .and. just_rhs .and. maxval(degree)==1 .and. &
       .not. resid_prob) then
      lrhs = grid%element(elem)%tags(chargetag)*chargeUnit/4
      go to 100
   endif
endif

! compute element quadrature points and weights

if (use_cached_basis) then
   select case (global_element_kind)
   case (TRIANGULAR_ELEMENT)
      nquad_pts = basis_at_quad_pts_tri_nqpoints(qdegree)
   case (TETRAHEDRAL_ELEMENT)
      nquad_pts = basis_at_quad_pts_tet_nqpoints(qdegree)
   end select
else
   select case (global_element_kind)
   case (TRIANGULAR_ELEMENT)
      call quadrature_rule_tri(qdegree,xvert,yvert,nquad_pts,quad_weight, &
                               xquad,yquad,jerr,stay_in=.true.)
      allocate(zquad(nquad_pts),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("allocation failed in elemental_matrix")
         stop
      endif
      zquad = 0.0_my_real
   case (TETRAHEDRAL_ELEMENT)
      call quadrature_rule_tet(qdegree,xvert,yvert,zvert,nquad_pts, &
                               quad_weight,xquad,yquad,zquad,jerr)
   end select
   if (jerr /= 0) then
      select case(jerr)
      case (1)
         ierr = PHAML_INTERNAL_ERROR
         call fatal("No rule for quadrature of requested order.",intlist=(/qdegree/))
         stop
      case(2)
         ierr = ALLOC_FAILED
         call fatal("allocation failed in quadrature_rules")
         stop
      case default
         ierr = PHAML_INTERNAL_ERROR
         call fatal("Quadrature rule returned error ",intlist=(/jerr/))
         stop
      end select
   endif
endif

! compute the values of the basis functions at the quadrature points

if (.not. allocated(basis)) then
   allocate(basis(nbasis,nquad_pts),basisx(nbasis,nquad_pts), &
            basisy(nbasis,nquad_pts),basisz(nbasis,nquad_pts),stat=astat)
elseif (size(basis,dim=1) /= nbasis .or. size(basis,dim=2) /= nquad_pts) then
   deallocate(basis,basisx,basisy,basisz)
   allocate(basis(nbasis,nquad_pts),basisx(nbasis,nquad_pts), &
            basisy(nbasis,nquad_pts),basisz(nbasis,nquad_pts),stat=astat)
else
   astat = 0
endif
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in elemental_matrix")
   return
endif

select case (global_element_kind)
case (TRIANGULAR_ELEMENT)
   select case (which_basis)
   case("n")
      call nodal_basis_func(xquad,yquad,xvert,yvert,degree(1),which_set, &
                            basis,basisx,basisy)
   case("h")
      call h_hier_basis_func(xquad,yquad,xvert,yvert,degree(1),which_set, &
                             basis,basisx,basisy)
   case("p")
      if (resid_prob .or. (not_just_rhs .and. .not.use_laplacian_table)) then
         if (use_cached_basis) then
            call p_hier_basis_at_quad_pts_tri(qdegree,degree,xvert,yvert,basis,&
                                              basisx,basisy,nquad_pts, &
                                              quad_weight,xquad,yquad)
            allocate(zquad(nquad_pts),stat=astat)
            if (astat /= 0) then
               ierr = ALLOC_FAILED
               call fatal("allocation failed in elemental_matrix")
               stop
            endif
            zquad = 0.0_my_real
         else
            call p_hier_basis_func(xquad,yquad,xvert,yvert,degree,which_set, &
                                   basis,basisx,basisy)
         endif
      else
         if (use_cached_basis) then
            call p_hier_basis_at_quad_pts_tri(qdegree,degree,xvert,yvert,basis,&
                                              nqpoints=nquad_pts, &
                                              qweights=quad_weight,xquad=xquad,&
                                              yquad=yquad)
            allocate(zquad(nquad_pts),stat=astat)
            if (astat /= 0) then
               ierr = ALLOC_FAILED
               call fatal("allocation failed in elemental_matrix")
               stop
            endif
            zquad = 0.0_my_real
         else
            call p_hier_basis_func(xquad,yquad,xvert,yvert,degree,which_set, &
                                   basis)
         endif
      endif
   end select
case (TETRAHEDRAL_ELEMENT)
   select case (which_basis)
   case("n")
      call nodal_basis_func(xquad,yquad,zquad,xvert,yvert,zvert,degree(1), &
                            which_set,basis,basisx,basisy,basisz)
   case("h")
      call h_hier_basis_func(xquad,yquad,zquad,xvert,yvert,zvert,degree(1), &
                             which_set,basis,basisx,basisy,basisz)
   case("p")
      if (not_just_rhs) then
         if (use_cached_basis) then
            call p_hier_basis_at_quad_pts_tet(qdegree,degree,xvert,yvert,zvert,&
                                              basis,basisx,basisy,basisz, &
                                              nquad_pts,quad_weight,xquad, &
                                              yquad,zquad)
         else
            call p_hier_basis_func(xquad,yquad,zquad,xvert,yvert,zvert,degree, &
                                   which_set,basis,basisx,basisy,basisz)
         endif
      else
         if (use_cached_basis) then
            call p_hier_basis_at_quad_pts_tet(qdegree,degree,xvert,yvert,zvert,&
                                              basis,nqpoints=nquad_pts, &
                                              qweights=quad_weight,xquad=xquad,&
                                              yquad=yquad,zquad=zquad)
         else
            call p_hier_basis_func(xquad,yquad,zquad,xvert,yvert,zvert,degree, &
                                   which_set,basis)
         endif
      endif
   end select
end select


! initialize sums to 0.0

if (not_just_rhs) then
   lmat = 0.0_my_real
   if (present(lmassmat)) lmassmat = 0.0_my_real
   if (present(extra_rhs)) extra_rhs = 0.0_my_real
   if (present(llapmat)) llapmat = 0.0_my_real
endif
lrhs = 0.0_my_real

if (present(extra_rhs)) then
   edim = 1+size(extra_rhs,2)
else
   edim = 1
endif

! for residual problems, evaluate the solution and derivatives, and for
! eigenvalue problems, evaluate the solution

if (resid_prob) then
   astat = 0
   astat1 = 0
   if (.not. allocated(u)) then
      allocate(u(ss,edim,nquad_pts),ux(ss,edim,nquad_pts), &
               uy(ss,edim,nquad_pts),uz(ss,edim,nquad_pts),stat=astat)
   else
      if (size(u,dim=1) /= ss .or. size(u,dim=2) /= edim .or. &
          size(u,dim=3) /= nquad_pts) then
         deallocate(u)
         allocate(u(ss,edim,nquad_pts),stat=astat1)
      endif
      if (.not. allocated(ux)) then
         allocate(ux(ss,edim,nquad_pts),uy(ss,edim,nquad_pts), &
                  uz(ss,edim,nquad_pts),stat=astat)
      elseif (size(ux,dim=1) /= ss .or. size(ux,dim=2) /= edim .or. &
          size(ux,dim=3) /= nquad_pts) then
         deallocate(ux,uy,uz)
         allocate(ux(ss,edim,nquad_pts),uy(ss,edim,nquad_pts), &
                  uz(ss,edim,nquad_pts),stat=astat)
      else
         astat = 0
      endif
   endif
   if (astat /= 0 .or. astat1 /= 0) then
      ierr = ALLOC_FAILED
      call fatal("allocation failed in elemental_matrix")
      return
   endif
   select case (global_element_kind)
   case (TRIANGULAR_ELEMENT)
      call evaluate_soln_local(grid,xquad,yquad,elem,(/(i,i=1,ss)/), &
                               (/(j,j=1,edim)/),u,ux,uy)
   case (TETRAHEDRAL_ELEMENT)
      call evaluate_soln_local(grid,xquad,yquad,elem,(/(i,i=1,ss)/), &
                               (/(j,j=1,edim)/),u,ux,uy,z=zquad,uz=uz)
   end select

elseif (grid%num_eval > 0) then
   if (.not. allocated(u)) then
      allocate(u(ss,edim,nquad_pts),stat=astat)
   elseif (size(u,dim=1) /= ss .or. size(u,dim=2) /= edim .or. &
       size(u,dim=3) /= nquad_pts) then
      deallocate(u)
      allocate(u(ss,edim,nquad_pts),stat=astat)
   else
      astat = 0
   endif
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("allocation failed in elemental_matrix")
      return
   endif
   select case (global_element_kind)
   case (TRIANGULAR_ELEMENT)
      call evaluate_soln_local(grid,xquad,yquad,elem,(/(i,i=1,ss)/), &
                               (/(j,j=1,edim)/),u)
   case (TETRAHEDRAL_ELEMENT)
      call evaluate_soln_local(grid,xquad,yquad,elem,(/(i,i=1,ss)/), &
                               (/(j,j=1,edim)/),u,z=zquad)
   end select
endif

if (use_laplacian_table .and. not_just_rhs) then
   call laplace_elemental_matrix(degree,lmat)
   if (present(llapmat)) llapmat = lmat
endif

! for each quadrature point

do qp = 1,nquad_pts

! evaluate pde coefficients

   call my_pdecoefs(elem,xquad(qp),yquad(qp),zquad(qp),cxx,cyy,czz,cxy,cxz, &
                    cyz,cx,cy,cz,c,rs)

   if (ss == 1) then
      if (.not. use_laplacian_table .and. not_just_rhs) then

! special BLAS based code for scalar PDEs

         if (my_real == kind(1.0)) then
            call sgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*cxx(1,1), &
                       basisx(1,qp),nbasis,basisx(1,qp),1,1.0,lmat,size(lmat,1))
            call sgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*cyy(1,1), &
                       basisy(1,qp),nbasis,basisy(1,qp),1,1.0,lmat,size(lmat,1))
            if (include_cross) then
               if (cxy(1,1) /= 0.0_my_real) then
                  call sgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*cxy(1,1), &
                             basisx(1,qp),nbasis,basisy(1,qp),1,1.0,lmat, &
                             size(lmat,1))
               endif
            endif
            if (include_first) then
               if (cx(1,1) /= 0.0_my_real) then
                  call sgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*cx(1,1), &
                             basis(1,qp),nbasis,basisx(1,qp),1,1.0,lmat, &
                             size(lmat,1))
               endif
               if (cy(1,1) /= 0.0_my_real) then
                  call sgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*cy(1,1), &
                             basis(1,qp),nbasis,basisy(1,qp),1,1.0,lmat, &
                             size(lmat,1))
               endif
            endif
            if (c(1,1) /= 0.0_my_real) then
               call sgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*c(1,1), &
                       basis(1,qp),nbasis,basis(1,qp),1,1.0,lmat,size(lmat,1))
            endif
            if (global_element_kind == TETRAHEDRAL_ELEMENT) then
               call sgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*czz(1,1), &
                       basisz(1,qp),nbasis,basisz(1,qp),1,1.0,lmat,size(lmat,1))
               if (include_cross) then
                  if (cxz(1,1) /= 0.0_my_real) then
                     call sgemm("N","N",nbasis,nbasis,1, &
                                quad_weight(qp)*cxz(1,1),basisx(1,qp),nbasis, &
                                basisz(1,qp),1,1.0,lmat,size(lmat,1))
                  endif
                  if (cyz(1,1) /= 0.0_my_real) then
                     call sgemm("N","N",nbasis,nbasis,1, &
                                quad_weight(qp)*cyz(1,1),basisy(1,qp),nbasis, &
                                basisz(1,qp),1,1.0,lmat,size(lmat,1))
                  endif
               endif
               if (include_first) then
                  if (cz(1,1) /= 0.0_my_real) then
                     call sgemm("N","N",nbasis,nbasis,1, &
                                quad_weight(qp)*cz(1,1),basis(1,qp),nbasis, &
                                basisz(1,qp),1,1.0,lmat,size(lmat,1))
                  endif
               endif
            endif
            if (present(lmassmat)) then
               call sgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*rs(1), &
                          basis(1,qp),nbasis,basis(1,qp),1,1.0,lmassmat, &
                          size(lmassmat,1))
            endif
            if (present(llapmat)) then
               call sgemm("N","N",nbasis,nbasis,1,quad_weight(qp), &
                          basisx(1,qp),nbasis,basisx(1,qp),1,1.0,llapmat, &
                          size(lmat,1))
               call sgemm("N","N",nbasis,nbasis,1,quad_weight(qp), &
                          basisy(1,qp),nbasis,basisy(1,qp),1,1.0,llapmat, &
                          size(lmat,1))
            endif
         elseif (my_real == kind(1.0d0)) then
            call dgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*cxx(1,1), &
                    basisx(1,qp),nbasis,basisx(1,qp),1,1.0d0,lmat,size(lmat,1))
            call dgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*cyy(1,1), &
                    basisy(1,qp),nbasis,basisy(1,qp),1,1.0d0,lmat,size(lmat,1))
            if (include_cross) then
               if (cxy(1,1) /= 0.0_my_real) then
                  call dgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*cxy(1,1), &
                             basisx(1,qp),nbasis,basisy(1,qp),1,1.0d0,lmat, &
                             size(lmat,1))
               endif
            endif
            if (include_first) then
               if (cx(1,1) /= 0.0_my_real) then
                  call dgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*cx(1,1), &
                             basis(1,qp),nbasis,basisx(1,qp),1,1.0d0,lmat, &
                             size(lmat,1))
               endif
               if (cy(1,1) /= 0.0_my_real) then
                  call dgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*cy(1,1), &
                             basis(1,qp),nbasis,basisy(1,qp),1,1.0d0,lmat, &
                             size(lmat,1))
               endif
            endif
            if (c(1,1) /= 0.0_my_real) then
               call dgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*c(1,1), &
                       basis(1,qp),nbasis,basis(1,qp),1,1.0d0,lmat,size(lmat,1))
            endif
            if (global_element_kind == TETRAHEDRAL_ELEMENT) then
               call dgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*czz(1,1), &
                     basisz(1,qp),nbasis,basisz(1,qp),1,1.0d0,lmat,size(lmat,1))
               if (include_cross) then
                  if (cxz(1,1) /= 0.0_my_real) then
                     call dgemm("N","N",nbasis,nbasis,1, &
                                quad_weight(qp)*cxz(1,1),basisx(1,qp),nbasis, &
                                basisz(1,qp),1,1.0d0,lmat,size(lmat,1))
                  endif
                  if (cyz(1,1) /= 0.0_my_real) then
                     call dgemm("N","N",nbasis,nbasis,1, &
                                quad_weight(qp)*cyz(1,1),basisy(1,qp),nbasis, &
                                basisz(1,qp),1,1.0d0,lmat,size(lmat,1))
                  endif
               endif
               if (include_first) then
                  if (cz(1,1) /= 0.0_my_real) then
                     call dgemm("N","N",nbasis,nbasis,1, &
                                quad_weight(qp)*cz(1,1),basis(1,qp),nbasis, &
                                basisz(1,qp),1,1.0d0,lmat,size(lmat,1))
                  endif
               endif
            endif
            if (present(lmassmat)) then
               call dgemm("N","N",nbasis,nbasis,1,quad_weight(qp)*rs(1), &
                          basis(1,qp),nbasis,basis(1,qp),1,1.0d0,lmassmat, &
                          size(lmassmat,1))
            endif
            if (present(llapmat)) then
               call dgemm("N","N",nbasis,nbasis,1,quad_weight(qp), &
                          basisx(1,qp),nbasis,basisx(1,qp),1,1.0d0,llapmat, &
                          size(lmat,1))
               call dgemm("N","N",nbasis,nbasis,1,quad_weight(qp), &
                          basisy(1,qp),nbasis,basisy(1,qp),1,1.0d0,llapmat, &
                          size(lmat,1))
            endif
         else
            ierr = PHAML_INTERNAL_ERROR
            call fatal("my_real is neither single nor double precision. Can't call GEMM")
            stop 
         endif
      endif
      do j=1,nbasis
         if (grid%num_eval <= 0) then
            lrhs(j) = lrhs(j) + quad_weight(qp) * rs(1)*basis(j,qp)
         else
            lrhs(j) = lrhs(j) + quad_weight(qp) * &
               rs(1)*grid%eigen_results%eigenvalue(1)*u(1,1,qp)*basis(j,qp)
         endif
         if (present(extra_rhs)) then
            do i=1,size(extra_rhs,2)
               extra_rhs(j,i) = extra_rhs(j,i) + quad_weight(qp) * rs(1) * &
                  grid%eigen_results%eigenvalue(i+1)*u(1,i+1,qp)*basis(j,qp)
            end do
         endif
         if (resid_prob) then
            term = -cxx(1,1)*ux(1,1,qp)*basisx(j,qp) &
                   -cyy(1,1)*uy(1,1,qp)*basisy(j,qp) &
                     -c(1,1)*u(1,1,qp)*basis(j,qp)
            if (include_cross) then
               term = term -cxy(1,1)*uy(1,1,qp)*basisx(j,qp)
            endif
            if (include_first) then
               term = term - cx(1,1)*ux(1,1,qp)*basis(j,qp) &
                           - cy(1,1)*uy(1,1,qp)*basis(j,qp)
            endif
            if (global_element_kind == TETRAHEDRAL_ELEMENT) then
               term = term -czz(1,1)*uz(1,1,qp)*basisz(j,qp)
               if (include_cross) then
                  term = term -cxz(1,1)*uz(1,1,qp)*basisx(j,qp) &
                              -cyz(1,1)*uz(1,1,qp)*basisy(j,qp)
               endif
               if (include_first) then
                  term = term - cz(1,1)*uz(1,1,qp)*basis(j,qp)
               endif
            endif
            lrhs(j) = lrhs(j) + quad_weight(qp)*term
            if (present(extra_rhs)) then
               do i=1,size(extra_rhs,2)
                  term = -cxx(1,1)*ux(1,i+1,qp)*basisx(j,qp) &
                         -cyy(1,1)*uy(1,i+1,qp)*basisy(j,qp) &
                           -c(1,1)*u(1,i+1,qp)*basis(j,qp)
                  if (include_cross) then
                     term = term -cxy(1,1)*uy(1,i+1,qp)*basisx(j,qp)
                  endif
                  if (include_first) then
                     term = term - cx(1,1)*ux(1,i+1,qp)*basis(j,qp) &
                                 - cy(1,1)*uy(1,i+1,qp)*basis(j,qp)
                  endif
                  if (global_element_kind == TETRAHEDRAL_ELEMENT) then
                     term = term -czz(1,1)*uz(1,i+1,qp)*basisz(j,qp)
                     if (include_cross) then
                        term = term -cxz(1,1)*uz(1,i+1,qp)*basisx(j,qp) &
                                    -cyz(1,1)*uz(1,i+1,qp)*basisy(j,qp)
                     endif
                     if (include_first) then
                        term = term - cz(1,1)*uz(1,i+1,qp)*basis(j,qp)
                     endif
                  endif
                  extra_rhs(j,i) = extra_rhs(j,i) + quad_weight(qp)*term
               end do
            endif
         endif
      end do

   else ! ss /= 1

! for each basis pair

      do j=1,nbasis
       if (not_just_rhs) then
        do i=1,nbasis

! contribution of this quadrature point to this integrals

          do l=1,ss
            do k=1,ss
               term = cxx(k,l)*basisx(i,qp)*basisx(j,qp) + &
                      cyy(k,l)*basisy(i,qp)*basisy(j,qp) + &
                        c(k,l)*basis (i,qp)*basis (j,qp)
               if (include_cross) then
                  term = term + cxy(k,l)*basisx(i,qp)*basisy(j,qp)
               endif
               if (include_first) then
                  term = term + cx(k,l)*basis (i,qp)*basisx(j,qp) + &
                                cy(k,l)*basis (i,qp)*basisy(j,qp)
               endif
               if (global_element_kind == TETRAHEDRAL_ELEMENT) then
                  term = term + czz(k,l)*basisz(i,qp)*basisz(j,qp)
                  if (include_cross) then
                     term = term + cxz(k,l)*basisx(i,qp)*basisz(j,qp) + &
                                   cyz(k,l)*basisy(i,qp)*basisz(j,qp)
                  endif
                  if (include_first) then
                     term = term + cz(k,l)*basis (i,qp)*basisz(j,qp)
                  endif 
               endif
               lmat(ss*(i-1)+k,ss*(j-1)+l) = lmat(ss*(i-1)+k,ss*(j-1)+l) + &
                                             quad_weight(qp)*term
               if (present(lmassmat) .and. k==l) then
                  lmassmat(ss*(i-1)+k,ss*(j-1)+l) = &
                                  lmassmat(ss*(i-1)+k,ss*(j-1)+l) + &
                                  quad_weight(qp)*basis(i,qp)*basis(j,qp)*rs(k)
               endif
               if (present(llapmat) .and. k==l) then
                  llapmat(ss*(i-1)+k,ss*(j-1)+l) = &
                                  llapmat(ss*(i-1)+k,ss*(j-1)+l) + &
                                  quad_weight(qp)*(basisx(i,qp)*basisx(j,qp) + &
                                                   basisy(i,qp)*basisy(j,qp))
               endif
            end do ! k
          end do ! l
        end do ! i
       endif ! not just rhs

! right hand side integral

       do k=1,ss
         if (grid%num_eval <= 0) then
             lrhs(ss*(j-1)+k) = lrhs(ss*(j-1)+k) + &
                                   quad_weight(qp) * rs(k)*basis(j,qp)
         else
             lrhs(ss*(j-1)+k) = lrhs(ss*(j-1)+k) + rs(k)* quad_weight(qp) * &
                grid%eigen_results%eigenvalue(1)*u(k,1,qp)*basis(j,qp)
         endif
         if (present(extra_rhs)) then
            do i=1,size(extra_rhs,2)
               extra_rhs(ss*(j-1)+k,i) = extra_rhs(ss*(j-1)+k,i) + &
                  quad_weight(qp) * rs(k) * &
                  grid%eigen_results%eigenvalue(i+1)*u(k,i+1,qp)*basis(j,qp)
            end do
         endif

! if setting up a residual problem, residual part of right hand side

          if (resid_prob) then
             do l=1,ss
                term = -cxx(k,l)*ux(l,1,qp)*basisx(j,qp) &
                       -cyy(k,l)*uy(l,1,qp)*basisy(j,qp) &
                         -c(k,l)*u(l,1,qp)*basis(j,qp)
                if (include_cross) then
                   term = term -cxy(k,l)*uy(l,1,qp)*basisx(j,qp)
                endif
                if (include_first) then
                   term = term - cx(k,l)*ux(l,1,qp)*basis(j,qp) &
                               - cy(k,l)*uy(l,1,qp)*basis(j,qp)
                endif
                if (global_element_kind == TETRAHEDRAL_ELEMENT) then
                   term = term - czz(k,l)*uz(l,1,qp)*basisz(j,qp)
                   if (include_cross) then
                      term = term - cxz(k,l)*uz(l,1,qp)*basisx(j,qp) &
                                  - cyz(k,l)*uz(l,1,qp)*basisy(j,qp)
                   endif
                   if (include_first) then
                      term = term - cz(k,l)*uz(l,1,qp)*basis(j,qp)
                   endif
                endif
                lrhs(ss*(j-1)+k) = lrhs(ss*(j-1)+k) + quad_weight(qp)*term
             end do
             if (present(extra_rhs)) then
                do i=1,size(extra_rhs,2)
                   do l=1,ss
                      term = -cxx(k,l)*ux(l,i+1,qp)*basisx(j,qp) &
                             -cyy(k,l)*uy(l,i+1,qp)*basisy(j,qp) &
                               -c(k,l)*u(l,i+1,qp)*basis(j,qp)
                      if (include_cross) then
                         term = term - cxy(k,l)*uy(l,i+1,qp)*basisx(j,qp)
                      endif
                      if (include_first) then
                         term = term - cx(k,l)*ux(l,i+1,qp)*basis(j,qp) &
                                     - cy(k,l)*uy(l,i+1,qp)*basis(j,qp)
                      endif
                      if (global_element_kind == TETRAHEDRAL_ELEMENT) then
                         term = term - czz(k,l)*uz(l,i+1,qp)*basisz(j,qp)
                         if (include_cross) then
                            term = term - cxz(k,l)*uz(l,i+1,qp)*basisx(j,qp) &
                                        - cyz(k,l)*uz(l,i+1,qp)*basisy(j,qp)
                         endif
                         if (include_first) then
                            term = term - cz(k,l)*uz(l,i+1,qp)*basis(j,qp)
                         endif
                      endif
                      extra_rhs(ss*(j-1)+k,i) = extra_rhs(ss*(j-1)+k,i) + &
                              quad_weight(qp)*term
                   end do
                end do
             endif
          endif
       end do
     end do

   endif ! ss == 1

end do

deallocate(quad_weight,xquad,yquad,zquad)

100 continue ! For special SEM rhs only, jump to here to get Neumann boundary

! compute boundary integrals

! experiments showed some cases needed one more degree of quadrature

qdegree = min(MAX_QUAD_ORDER_LINE,qdegree+1)

select case (global_element_kind)
case (TRIANGULAR_ELEMENT)
   ilimit = EDGES_PER_ELEMENT
   add2side = 0
case (TETRAHEDRAL_ELEMENT)
   ilimit = FACES_PER_ELEMENT
   add2side = EDGES_PER_ELEMENT
end select

do side=1,ilimit

! do sides that have at least one component that is NATURAL or MIXED

   doit = .false.
   do j=1,ss
      select case (global_element_kind)
      case (TRIANGULAR_ELEMENT)
         if (edge_type(side,j) == NATURAL .or. edge_type(side,j) == MIXED) then
            doit = .true.
            exit
         endif
      case (TETRAHEDRAL_ELEMENT)
         if (face_type(side,j) == NATURAL .or. face_type(side,j) == MIXED) then
            doit = .true.
            exit
         endif
      end select
   end do

   if (.not. doit) cycle

! compute side quadrature points and weights

   select case (side)
   case (1)
      xelem(1) = xvert(2)
      xelem(2) = xvert(3)
      yelem(1) = yvert(2)
      yelem(2) = yvert(3)
      if (global_element_kind == TETRAHEDRAL_ELEMENT) then
         xelem(3) = xvert(4)
         yelem(3) = yvert(4)
         zelem(1) = zvert(2)
         zelem(2) = zvert(3)
         zelem(3) = zvert(4)
      endif
   case (2)
      xelem(1) = xvert(3)
      xelem(2) = xvert(1)
      yelem(1) = yvert(3)
      yelem(2) = yvert(1)
      if (global_element_kind == TETRAHEDRAL_ELEMENT) then
         xelem(3) = xvert(4)
         yelem(3) = yvert(4)
         zelem(1) = zvert(3)
         zelem(2) = zvert(1)
         zelem(3) = zvert(4)
      endif
   case (3)
      xelem(1) = xvert(1)
      xelem(2) = xvert(2)
      yelem(1) = yvert(1)
      yelem(2) = yvert(2)
      if (global_element_kind == TETRAHEDRAL_ELEMENT) then
         xelem(3) = xvert(4)
         yelem(3) = yvert(4)
         zelem(1) = zvert(1)
         zelem(2) = zvert(2)
         zelem(3) = zvert(4)
      endif
   case (4)
      xelem(1) = xvert(1)
      xelem(2) = xvert(2)
      xelem(3) = xvert(3)
      yelem(1) = yvert(1)
      yelem(2) = yvert(2)
      yelem(3) = yvert(3)
      zelem(1) = zvert(1)
      zelem(2) = zvert(2)
      zelem(3) = zvert(3)
   end select
   select case (global_element_kind)
   case (TRIANGULAR_ELEMENT)
      call quadrature_rule_line(qdegree,xelem,yelem,nquad_pts,quad_weight, &
                                xquad,yquad,jerr)
      allocate(zquad(nquad_pts),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("allocation failed in elemental_matrix")
         stop
      endif
      zquad = 0.0_my_real
   case (TETRAHEDRAL_ELEMENT)
      call quadrature_rule_tri(qdegree,xelem,yelem,nquad_pts,quad_weight,xquad,&
                               yquad,jerr,zvert=zelem,zquad=zquad)
   end select
   if (jerr /= 0) then
      select case(jerr)
      case (1)
         ierr = PHAML_INTERNAL_ERROR
         call fatal("No rule for quadrature of requested order.", &
                    intlist=(/qdegree/))
         stop
      case(2)
         ierr = ALLOC_FAILED
         call fatal("allocation failed in quadrature_rules")
         stop
      case default
         ierr = PHAML_INTERNAL_ERROR
         call fatal("Quadrature rule returned error ",intlist=(/jerr/))
         stop
      end select
   endif

! set the values of the basis functions at the quadrature points

   if (.not. allocated(basis_bound)) then
      allocate(basis_bound(nbasis,nquad_pts),stat=astat)
   elseif (size(basis_bound,dim=1) /= nbasis .or. &
           size(basis_bound,dim=2) /= nquad_pts) then
      deallocate(basis_bound)
      allocate(basis_bound(nbasis,nquad_pts),stat=astat)
   else
      astat = 0
   endif
   if (.not. allocated(bcrs)) then
      allocate(bcrs(ss,nquad_pts),bcc(ss,ss,nquad_pts),stat=astat1)
   elseif (size(bcrs,dim=1) /= ss .or. size(bcrs,dim=2) /= nquad_pts) then
      deallocate(bcrs,bcc)
      allocate(bcrs(ss,nquad_pts),bcc(ss,ss,nquad_pts),stat=astat1)
   else
      astat1 = 0
   endif
   if (present(extra_rhs)) then
      if (.not. allocated(extra_bcrs)) then
         allocate(extra_bcrs(ss,size(extra_rhs,dim=2),nquad_pts),stat=astat2)
      elseif (size(extra_bcrs,dim=1) /= ss .or. &
              size(extra_bcrs,dim=2) /= size(extra_rhs,dim=2) .or. &
              size(extra_bcrs,dim=3) /= nquad_pts) then
         deallocate(extra_bcrs)
         allocate(extra_bcrs(ss,size(extra_rhs,dim=2),nquad_pts),stat=astat2)
      else
         astat2 = 0
      endif
   else
      astat2 = 0
   endif
   if (astat /= 0 .or. astat1 /= 0 .or. astat2 /= 0) then
      ierr = ALLOC_FAILED
      call fatal("allocation failed in elemental_matrix")
      return
   endif

   select case (global_element_kind)
   case (TRIANGULAR_ELEMENT)
      select case(which_basis)
      case("n")
         call nodal_basis_func(xquad,yquad,xvert,yvert,degree(1),which_set, &
                               basis_bound)
      case("h")
         call h_hier_basis_func(xquad,yquad,xvert,yvert,degree(1),which_set, &
                                basis_bound)
      case("p")
         call p_hier_basis_func(xquad,yquad,xvert,yvert,degree,which_set, &
                                basis_bound)
      end select
   case (TETRAHEDRAL_ELEMENT)
      select case(which_basis)
      case("n")
         call nodal_basis_func(xquad,yquad,zquad,xvert,yvert,zvert,degree(1), &
                               which_set,basis_bound)
      case("h")
         call h_hier_basis_func(xquad,yquad,zquad,xvert,yvert,zvert,degree(1), &
                               which_set,basis_bound)
      case("p")
         call p_hier_basis_func(xquad,yquad,zquad,xvert,yvert,zvert,degree, &
                               which_set,basis_bound)
      end select
   end select

! evaluate boundary conditions at all quadrature points

   secret_elem = elem
   if (present(loc_bconds_a)) then
      if (present(extra_rhs)) then
         call loc_bconds_a(xquad,yquad,zquad,bmark(side+add2side),bctype,bcc, &
                           bcrs,extra_bcrs)
      else
         call loc_bconds_a(xquad,yquad,zquad,bmark(side+add2side),bctype,bcc, &
                           bcrs)
      endif
   elseif (present(loc_bconds_s)) then
      do qp=1,nquad_pts
         call loc_bconds_s(xquad(qp),yquad(qp),zquad(qp),bmark(side+add2side), &
                           bctype,bcc(:,:,qp),bcrs(:,qp))
         if (present(extra_rhs)) then
            do i=1,size(extra_bcrs,dim=2)
               extra_bcrs(:,i,qp) = bcrs(:,qp)
            end do
         endif
      end do
   else
      ierr = PHAML_INTERNAL_ERROR
      call fatal("must give one of loc_bconds forms in elemental_matrix")
      stop
   endif

! for each quadrature point

   do qp=1,nquad_pts

! for each pair of basis functions

      do j=1,nbasis
         if (not_just_rhs) then
            do i=1,nbasis
               do l=1,ss
                  do k=1,ss
                     if (bctype(k) /= MIXED) cycle
                     lmat(ss*(i-1)+k,ss*(j-1)+l) = &
                        lmat(ss*(i-1)+k,ss*(j-1)+l) + &
                        quad_weight(qp)*bcc(k,l,qp)* &
                        basis_bound(i,qp)*basis_bound(j,qp)
                  end do
               end do
            end do
         endif
         do k=1,ss
            if (bctype(k) /= NATURAL .and. bctype(k) /= MIXED) cycle
            lrhs(ss*(j-1)+k) = lrhs(ss*(j-1)+k) + &
                                    quad_weight(qp)*bcrs(k,qp)*basis_bound(j,qp)
            if (present(extra_rhs)) then
               do i=1,size(extra_rhs,2)
                  extra_rhs(ss*(j-1)+k,i) = extra_rhs(ss*(j-1)+k,i) + &
                            quad_weight(qp)*extra_bcrs(k,i,qp)*basis_bound(j,qp)
               end do
            endif
         end do
      end do
   end do

   deallocate(quad_weight,xquad,yquad,zquad)
   if (present(extra_rhs)) then
      deallocate(extra_bcrs)
   endif

end do

end subroutine elemental_matrix

!          --------
subroutine assemble(linear_system,eq_num,nleq,local_stiffness,local_rhs, &
                    local_mass)
!          --------

!----------------------------------------------------
! This routine adds the local matrix from one element to the global matrix.
! The equations for that element are given in eq_num.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linear_system
integer, intent(in) :: eq_num(:), nleq
real(my_real), intent(in) :: local_stiffness(:,:),local_rhs(:),local_mass(:,:)

!----------------------------------------------------
! Local variables:

integer :: eq, col, i, leq, lcol, syssize, eq1, eq2, s1, s2

!----------------------------------------------------
! Begin executable code

syssize = linear_system%system_size

! for each equation ...

do eq1=1,nleq
   do s1=1,syssize
      leq = syssize*(eq1-1)+s1
      eq = eq_num(eq1)+s1-1

! add contribution to right hand side

!$omp atomic
      linear_system%rhs(eq) = linear_system%rhs(eq) + local_rhs(leq)

      if (.not. linear_system%linsys_exists) then

! for each column

         do eq2=1,nleq
            do s2=1,syssize
               lcol = syssize*(eq2-1)+s2
               col = eq_num(eq2)+s2-1

! find the column in the global matrix

               i = find_col_in_row(linear_system,eq,col)
               if (i == 0) then
                  ierr = PHAML_INTERNAL_ERROR
                  call fatal("failed to find column in assemble")
                  stop
               endif

! add contribution to the matrix

!$omp atomic
               linear_system%stiffness(i) = linear_system%stiffness(i) + &
                                            local_stiffness(leq,lcol)
               if (associated(linear_system%mass)) then
!$omp atomic
                  linear_system%mass(i) = linear_system%mass(i) + &
                                          local_mass(leq,lcol)
               endif

            end do ! next PDE
         end do ! next column
      endif ! linsys_exists
   end do ! next PDE
end do ! next equation

end subroutine assemble

!          --------------
subroutine fix_quad_error(grid,procs,linear_system,still_sequential,timeit)
!          --------------

!----------------------------------------------------
! This routine gets the contributions to the matrix and right hand side
! from elements owned by other processors.  This is used instead of
! computing them myself so that the quadrature errors are on the order of
! the fine grid elements instead of the element I have.
! This only works for linear elements
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
type(proc_info), intent(in) :: procs
type(linsys_type), intent(inout) :: linear_system
logical, intent(in) :: still_sequential, timeit

!----------------------------------------------------
! Local variables:

integer :: i, lev, eq, indx, allocstat, my_processor, nonzero, &
           nproc,p, j, k, col, elem, nisend, isub, rsub, sisub, srsub
logical :: rhs_next, mass_next
logical(small_logical) :: need_fix(linear_system%neq_vert)
type(hash_key_eq) :: gid
integer, allocatable :: send_int(:), nirecv(:), nrrecv(:),nisendv(:), nrsendv(:)
integer, pointer :: recv_int(:)
real(my_real), allocatable :: send_real(:)
real(my_real), pointer :: recv_real(:)

!----------------------------------------------------
! Begin executable code

nproc = num_proc(procs)
my_processor = my_proc(procs)

! master doesn't do this and nothing to do if sequential

if (my_processor == MASTER .or.  still_sequential .or. &
    nproc == 1) return

! don't need to fix quadrature error in 3D because hbmg multigrid is not
! implemented in 3D

if (global_element_kind == TETRAHEDRAL_ELEMENT) return

! some memory for messages

allocate(nirecv(nproc),nrrecv(nproc),nisendv(nproc),nrsendv(nproc), &
         stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in fix_quad_error",procs=procs)
   return
endif

! determine which equations need fixing

need_fix = .false.
elem=grid%head_level_elem(1)
do while (elem /= END_OF_LIST)
   call mark_need_fix(grid,linear_system,need_fix,elem)
   elem = grid%element(elem)%next
end do

! convert to the hierarchical basis

linear_system%matrix_val => linear_system%stiffness
do lev=linear_system%bubble_level,2,-1
   call basis_change(lev,TO_HIER,linear_system,do_mass=.true.)
end do

! count the number of equations that need fixing

nisend = 0
do eq=1,linear_system%neq_vert
   if (need_fix(eq)) then
      nisend = nisend + KEY_SIZE+1
   endif
end do

allocate(send_int(nisend),stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in fix_quad_error",procs=procs)
   return
endif

! make a list of the equations that need fixing

nisend = 0
do eq=1,linear_system%neq_vert
   if (need_fix(eq)) then
      call hash_pack_key(linear_system%gid(eq),send_int,nisend+1)
      nisend = nisend + KEY_SIZE+1
   endif
end do

! send the list

if (timeit) call start_watch((/cpassemble,ctassemble/))
call phaml_alltoall(procs,send_int,nisend,recv_int,nirecv,1102)
if (timeit) call stop_watch((/cpassemble,ctassemble/))

deallocate(send_int,stat=allocstat)

! count the number of entries to be sent to each processor

nisendv = 0
nrsendv = 0
isub = 1
do p=1,nproc
   if (p == my_processor) then
      isub = isub + nirecv(p)
      cycle
   endif
   do i=1,nirecv(p)/(KEY_SIZE+1)
      gid = hash_unpack_key(recv_int,isub,.true.)
      isub = isub + KEY_SIZE+1
      eq = hash_decode_key(gid,linear_system%eq_hash)
      if (eq /= HASH_NOT_FOUND) then

! how many nonzeros in the matrix row

         nonzero = 0
         if (.not. linear_system%linsys_exists) then
            do j=linear_system%begin_row(eq),linear_system%end_row_linear(eq)
               if (linear_system%column_index(j) == NO_ENTRY) cycle
               if (associated(linear_system%mass)) then
                  if (linear_system%stiffness(j) /= 0.0_my_real .or. &
                     linear_system%mass(j) /= 0.0_my_real) nonzero = nonzero + 2
               else
                  if (linear_system%stiffness(j) /= 0.0_my_real) then
                     nonzero = nonzero + 1
                  endif
               endif
            end do
         endif

! count the rhs if there are any nonzeros in the matrix or the rhs is nonzero,
! plus also one for a NULL_KEY marker to end the row

         if (nonzero > 0 .or. linear_system%rhs(eq) /= 0.0_my_real) then
            nisendv(p) = nisendv(p) + (nonzero+2)*(KEY_SIZE+1)
            nrsendv(p) = nrsendv(p) + nonzero+1
         else
            nisendv(p) = nisendv(p) + nonzero*(KEY_SIZE+1)
            nrsendv(p) = nrsendv(p) + nonzero
         endif

      endif
   end do
end do

! allocate memory for the messages

allocate(send_int(sum(nisendv)),send_real(sum(nrsendv)),stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in fix_quad_error",procs=procs)
   return
endif

! pack the data
! the data for each row consists of
!   first the equation gid in the int and rhs in the real
!   then the stiffness matrix and mass matrix values in the real and column
!   gid in the int whereever either the stiffness matrix or mass matrix
!   is nonzero, finally a NULL hash key to indicate end-of-row.

isub = 1
sisub = 1
srsub = 1
do p=1,nproc
   if (p == my_processor) then
      isub = isub + nirecv(p)
      cycle
   endif
   do i=1,nirecv(p)/(KEY_SIZE+1)
      gid = hash_unpack_key(recv_int,isub,.true.)
      isub = isub + KEY_SIZE+1
      eq = hash_decode_key(gid,linear_system%eq_hash)
      if (eq /= HASH_NOT_FOUND) then

! how many nonzeros in the matrix row

         nonzero = 0
         if (.not. linear_system%linsys_exists) then
            do j=linear_system%begin_row(eq),linear_system%end_row_linear(eq)
               if (linear_system%column_index(j) == NO_ENTRY) cycle
               if (associated(linear_system%mass)) then
                  if (linear_system%stiffness(j) /= 0.0_my_real .or. &
                      linear_system%mass(j) /= 0.0_my_real) then
                     nonzero = 2
                     exit
                  endif
               else
                  if (linear_system%stiffness(j) /= 0.0_my_real) then
                     nonzero = 1
                     exit
                  endif
               endif
            end do
         endif

! pack the data

         if (nonzero > 0 .or. linear_system%rhs(eq) /= 0.0_my_real) then
            send_int(sisub:sisub+KEY_SIZE) = recv_int(isub-(KEY_SIZE+1):isub-1)
            sisub = sisub + KEY_SIZE+1
            send_real(srsub) = linear_system%rhs(eq)
            srsub = srsub + 1
            if (.not. linear_system%linsys_exists) then
               do j=linear_system%begin_row(eq),linear_system%end_row_linear(eq)
                  if (linear_system%column_index(j) == NO_ENTRY) cycle
                  if (associated(linear_system%mass)) then
                     if (linear_system%stiffness(j) /= 0.0_my_real .or. &
                         linear_system%mass(j) /= 0.0_my_real) then
                        call hash_pack_key(linear_system%gid(linear_system%column_index(j)), &
                                           send_int,sisub)
                        sisub = sisub + KEY_SIZE+1
                        send_real(srsub) = linear_system%stiffness(j)
                        srsub = srsub + 1
                        call hash_pack_key(linear_system%gid(linear_system%column_index(j)), &
                                           send_int,sisub)
                        sisub = sisub + KEY_SIZE+1
                        send_real(srsub) = linear_system%mass(j)
                        srsub = srsub + 1
                     endif
                  else
                     if (linear_system%stiffness(j) /= 0.0_my_real) then
                        call hash_pack_key(linear_system%gid(linear_system%column_index(j)), &
                                           send_int,sisub)
                        sisub = sisub + KEY_SIZE+1
                        send_real(srsub) = linear_system%stiffness(j)
                        srsub = srsub + 1
                     endif
                  endif
               end do
            endif
            call hash_pack_key(NULL_KEY_EQ,send_int,sisub)
            sisub = sisub + KEY_SIZE+1
         endif
      endif
   end do
end do

if (associated(recv_int)) deallocate(recv_int,stat=allocstat)

! send the lists to the other processors

if (timeit) call start_watch((/cpassemble,ctassemble/))
call phaml_alltoall(procs,send_int,nisendv,recv_int,nirecv,1103)
call phaml_alltoall(procs,send_real,nrsendv,recv_real,nrrecv,1104)
if (timeit) call stop_watch((/cpassemble,ctassemble/))

deallocate(send_int,send_real,stat=allocstat)

! receive the lists and add contributions to my linear system

isub = 1
rsub = 1
do p=1,nproc
   if (p == my_processor) cycle
   rhs_next = .true.
   mass_next = .false.
   do j=1,nirecv(p)/(KEY_SIZE+1)
      gid = hash_unpack_key(recv_int,isub,.true.)
      isub = isub + KEY_SIZE+1
      if (gid == NULL_KEY_EQ) then
         rhs_next = .true.
      elseif (rhs_next) then
         rhs_next = .false.
         eq = hash_decode_key(gid,linear_system%eq_hash)
         if (eq == HASH_NOT_FOUND) then
            call warning("received data for equation I don't have in fix_quad_error")
            call hash_print_key(gid,errunit)
            cycle
         endif
         linear_system%rhs(eq) = linear_system%rhs(eq) + recv_real(rsub)
         rsub = rsub + 1
      elseif (mass_next) then
! indx was computed in the else clause of the previous time through the loop
         if (indx /= 0) then
            linear_system%mass(indx) = linear_system%mass(indx) +recv_real(rsub)
            rsub = rsub + 1
         endif
         mass_next = .false.
      else
         if (associated(linear_system%mass)) mass_next = .true.
         if (eq == HASH_NOT_FOUND) cycle
         col = hash_decode_key(gid,linear_system%eq_hash)
         if (col == HASH_NOT_FOUND) then
            call warning("received data for column I don't have in fix_quad_error")
            call hash_print_key(gid,errunit)
            cycle
         endif
! find the location of this column, if I have it
         indx = 0
         do k=linear_system%begin_row(eq),linear_system%end_row_linear(eq)
            if (linear_system%column_index(k) == col) then
               indx = k
               exit
            endif
         end do
         if (indx /= 0) then
            linear_system%stiffness(indx) = linear_system%stiffness(indx) + &
                                            recv_real(rsub)
            rsub = rsub + 1
         else
            call warning("couldn't find column in fix_quad_error")
         endif
      endif
   end do
end do

! convert back to nodal basis

do lev=2,linear_system%bubble_level
   call basis_change(lev,TO_NODAL,linear_system,do_mass=.true.)
end do

! free message memory

if (associated(recv_int)) deallocate(recv_int,stat=allocstat)
if (associated(recv_real)) deallocate(recv_real,stat=allocstat)
deallocate(nisendv,nrsendv,nirecv,nrrecv,stat=allocstat)

end subroutine fix_quad_error

!                    -------------
recursive subroutine mark_need_fix(grid,linear_system,need_fix,elem)
!                    -------------

!----------------------------------------------------
! This routine determines which equations need quadrature contributions
! from other processors
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
type(linsys_type), intent(in) :: linear_system
logical(small_logical), intent(inout) :: need_fix(:)
integer, intent(in) :: elem
!----------------------------------------------------
! Local variables:

integer :: i,j,eq,child(MAX_CHILD),allc(MAX_CHILD)
!----------------------------------------------------
! Begin executable code

! recursively traverse the refinement tree.  When I own an element, quit
! traversing that branch.  Otherwise, mark the equations associated with
! the three vertices as needing to be fixed.

if (grid%element(elem)%iown) return

do i=1,VERTICES_PER_ELEMENT
   do j=1,linear_system%system_size
      call grid_to_eq(grid,linear_system,VERTEX_ID,1,j, &
                      grid%vertex(grid%element(elem)%vertex(i))%gid,eq)
      need_fix(eq) = .true.
   end do
end do

allc = ALL_CHILDREN
child = get_child_lid(grid%element(elem)%gid,allc,grid%elem_hash)

if (child(1) == NO_CHILD) return

do i=1,MAX_CHILD
   call mark_need_fix(grid,linear_system,need_fix,child(i))
end do

end subroutine mark_need_fix

!          -------------------
subroutine static_condensation(linear_system,grid,procs)
!          -------------------

!----------------------------------------------------
! This routine performs static condensation to eliminate the bubble bases.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linear_system
type(grid_type), intent(in) :: grid
type(proc_info), intent(in) :: procs
!----------------------------------------------------
! Local variables:

integer :: i, j, k, row, col, eq, objtype, brank, srank, lid, degree, loc_neq, &
           syssize, rowlen, info, eq_list(linear_system%neq), iblock, nblock, &
           astat, min_bubble, max_bubble
real(my_real) :: temp
real(my_real), allocatable :: row_block(:,:), row_block_t(:,:)
real(my_real), pointer :: block(:,:)
integer, pointer :: ipiv(:)
!----------------------------------------------------
! Begin executable code

! if there are no bubble equations, just set the pointers and return

if (linear_system%begin_level(linear_system%bubble_level) == &
    linear_system%begin_level(linear_system%beyond_last_level)) then
   nullify(linear_system%elem_block)
   select case (global_element_kind)
   case (TRIANGULAR_ELEMENT)
      linear_system%end_row => linear_system%end_row_edge
   case (TETRAHEDRAL_ELEMENT)
      linear_system%end_row => linear_system%end_row_face
   end select
   linear_system%neq = linear_system%neq_cond
   linear_system%matrix_val => linear_system%condensed
   linear_system%rhs => linear_system%rhs_cond
   return
endif

syssize = linear_system%system_size

! factor the blocks of bubble bases from one element

! allocate the structure that contains the factored blocks

! go through the bubble equations, making a list of the first equation
! of each block corresponding to an element
! keep track of the minimum and maximum lids found to allocate elem_block

nblock = 0
min_bubble = huge(0)
max_bubble = -1
eq = linear_system%begin_level(linear_system%bubble_level)
do
   if (eq >= linear_system%begin_level(linear_system%beyond_last_level)) exit

! find the element corresponding to the block this row starts

   call eq_to_grid(linear_system,linear_system%gid(eq),objtype,brank,srank, &
                   lid,grid)

! make sure this is the beginning of a block

   if (brank /= 1 .or. srank /= 1) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("not at beginning of block in factorization of high order equations", &
                 intlist=(/eq,brank,srank,objtype,i/),procs=procs)
      stop
   endif

! only bubble equations should occur

   if (objtype /= ELEMENT_ID) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("object type invalid  when examining high order equation", &
                 intlist=(/objtype/),procs=procs)
      stop
   endif

   nblock = nblock + 1
   eq_list(nblock) = eq
   degree = grid%element(lid)%degree
   loc_neq = syssize*element_dof(degree)
   eq = eq + loc_neq
   min_bubble = min(min_bubble,lid)
   max_bubble = max(max_bubble,lid)

end do

allocate(linear_system%elem_block(min_bubble:max_bubble),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in static_condensation",procs=procs)
   return
endif
do i=min_bubble,max_bubble
   nullify(linear_system%elem_block(i)%matrix, &
           linear_system%elem_block(i)%ipiv)
   linear_system%elem_block(i)%neq = 0
end do
      
! check the kind of real for LAPACK

if (my_real /= kind(0.0) .and. my_real /= kind(0.0d0)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("lapack requires my_real is either default single or double precision")
   return
endif

! go through the element equation blocks

! OpenMP parallel section.  To get the same results as a sequential program,
! we need the ordered construct for multiplying by the Schur complement,
! because floating point arithmetic is not commutative.  To keep this
! somewhat efficient we need a chunk size of 1 (so threads aren't waiting
! for the previous thread to complete many loops), and since the amount of
! work varies strongly with the degree of the element, we use a dynamic
! schedule (which has a default chunk size of 1).

!$omp parallel do &
!$omp  ordered &
!$omp  schedule(dynamic) &
!$omp  default(shared) &
!$omp  private(iblock,eq,objtype,brank,srank,lid,degree,loc_neq,block,ipiv, &
!$omp          rowlen,row_block,row_block_t,i,j,col,info,row,temp,k)

do iblock = 1,nblock
   eq = eq_list(iblock)
   call eq_to_grid(linear_system,linear_system%gid(eq),objtype,brank,srank, &
                   lid,grid)

! get the element degree, determine number of equations, and allocate memory

   degree = grid%element(lid)%degree
   loc_neq = syssize*element_dof(degree)
         
   allocate(linear_system%elem_block(lid)%matrix(loc_neq,loc_neq), &
            linear_system%elem_block(lid)%ipiv(loc_neq))
   linear_system%elem_block(lid)%neq = loc_neq
   block => linear_system%elem_block(lid)%matrix
   ipiv => linear_system%elem_block(lid)%ipiv

! copy the entries for this element interior into block, and the rows and rhs
! into row_block and row_block_t (transpose)
! Dirichlet rows

   rowlen = linear_system%end_row(eq)-linear_system%begin_row(eq)+1
   allocate(row_block(loc_neq,rowlen+1),row_block_t(rowlen+1,loc_neq))
   row_block = 0.0_my_real
   row_block_t = 0.0_my_real

   block = 0.0_my_real
   do i=1,loc_neq
      do j=linear_system%begin_row(eq+i-1),linear_system%end_row(eq+i-1)
         col = linear_system%column_index(j)
         if (col >= eq .and. col < eq+loc_neq) then
            block(i,col-eq+1) = linear_system%condensed(j)
         endif
         row_block(i,j-linear_system%begin_row(eq+i-1)+1) = &
         linear_system%condensed(j)
         k = find_col_in_row(linear_system,col,eq+i-1)
         if (k == 0) then
            row_block_t(j-linear_system%begin_row(eq+i-1)+1,i) = 0.0_my_real
         else
            row_block_t(j-linear_system%begin_row(eq+i-1)+1,i) = &
               linear_system%condensed(k)
         endif
      end do
      row_block(i,rowlen+1) = linear_system%rhs(eq+i-1)
      row_block_t(rowlen+1,i) = linear_system%rhs(eq+i-1)
   end do

! factor the block to PLU

   if (my_real == kind(0.0)) then
      call sgetrf(loc_neq,loc_neq,block,loc_neq,ipiv,info)
   else
      call dgetrf(loc_neq,loc_neq,block,loc_neq,ipiv,info)
   endif
   if (info /= 0) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("lapack sgetrf failed for factoring high order bases", &
                 intlist=(/info/),procs=procs)
      stop
   endif

! form the Schur complement A1 - A21 A2^-1 A12 for the condensed equations
! (and also the rhs) where A1 is vertex and edge bases and A2 is bubble bases.

! multiply the rows and rhs by A2^-1

   if (my_real == kind(0.0)) then
      call sgetrs("N",loc_neq,rowlen+1,block,loc_neq,ipiv,row_block, &
                  loc_neq,info)
   else
      call dgetrs("N",loc_neq,rowlen+1,block,loc_neq,ipiv,row_block, &
                  loc_neq,info)
   endif
   if (info /= 0) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("lapack sgetrs solution failed for static condensation",&
                 intlist=(/info/),procs=procs)
      stop
   endif

! form the Schur complement

!$omp ordered
   do i=1,rowlen
      row = linear_system%column_index(linear_system%begin_row(eq)+i-1)
      if (row >= eq .and. row < eq+loc_neq) cycle
      do j=1,rowlen
         col = linear_system%column_index(linear_system%begin_row(eq)+j-1)
         if (col >= eq .and. col < eq+loc_neq) cycle
         temp = 0.0_my_real
         do k=1,loc_neq
            temp = temp + row_block_t(i,k)*row_block(k,j)
         end do
         k = find_col_in_row(linear_system,row,col)
         if (k == 0) then
            ierr = PHAML_INTERNAL_ERROR
            call fatal("couldn't find col in row during static condensation", &
                       intlist=(/col,row/),procs=procs)
            stop
         endif
         linear_system%condensed(k) = linear_system%condensed(k) - temp
      end do
! don't change the rhs of Dirichlet rows
      if (linear_system%equation_type(row) /= DIRICHLET) then
         temp = 0.0_my_real
         do k=1,loc_neq
            temp = temp + row_block_t(i,k)*row_block(k,rowlen+1)
         end do
         linear_system%rhs_cond(row) = linear_system%rhs_cond(row) - temp
      endif
   end do
!$omp end ordered

   deallocate(row_block,row_block_t)
end do ! blocks of bubble equations

!$omp end parallel do

! after static condensation, the matrix rows should not contain the bubble
! bases, the number of equations should not include bubble bases, and the
! matrix values and rhs should be the Schur complement

select case (global_element_kind)
case (TRIANGULAR_ELEMENT)
   linear_system%end_row => linear_system%end_row_edge
case (TETRAHEDRAL_ELEMENT)
   linear_system%end_row => linear_system%end_row_face
end select
linear_system%neq = linear_system%neq_cond
linear_system%matrix_val => linear_system%condensed
linear_system%rhs => linear_system%rhs_cond

end subroutine static_condensation

!          -----------------------
subroutine static_condensation_rhs(linear_system,grid,procs)
!          -----------------------

!----------------------------------------------------
! This routine performs static condensation on just the right hand side. The
! full matrix must have previously been condensed so the factored blocks exist.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout), target :: linear_system
type(grid_type), intent(in) :: grid
type(proc_info), intent(in) :: procs

!----------------------------------------------------
! Local variables:

integer :: eq, objtype, brank, srank, objlid, loc_neq, info, astat, &
           i, j, row, col, tvsize
real(my_real), pointer :: block(:,:)
integer, pointer :: ipiv(:)
real(my_real), allocatable :: tempvec(:,:)

!----------------------------------------------------
! Begin executable code

! Perform static condensation on rhs; b1 - A21 A2^-1 b2 where 2 is bubble bases

tvsize = element_dof(linear_system%maxdeg)
allocate(tempvec(tvsize,1),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in static_condensation_rhs")
   stop
endif

! For each block of equations corresponding to the bubble bases of an element...

eq = linear_system%begin_level(linear_system%bubble_level)
do
   if (eq >= linear_system%begin_level(linear_system%beyond_last_level)) exit
   call eq_to_grid(linear_system,linear_system%gid(eq),objtype,brank,srank, &
                   objlid,grid)

! block contains A2^-1

   block => linear_system%elem_block(objlid)%matrix
   ipiv => linear_system%elem_block(objlid)%ipiv
   loc_neq = linear_system%elem_block(objlid)%neq

   if (loc_neq > tvsize) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("static_condensation_rhs: insufficient size for tempvec")
      stop
   endif

! tempvec contains b2

   tempvec(1:loc_neq,1) = linear_system%rhs(eq:eq+loc_neq-1)

! Compute A2^-1 b2, result is in tempvec

   if (my_real == kind(0.0)) then
     call sgetrs("N",loc_neq,1,block,loc_neq,ipiv,tempvec,tvsize,info)
   else
     call dgetrs("N",loc_neq,1,block,loc_neq,ipiv,tempvec,tvsize,info)
   endif
   if (info /= 0) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("lapack sgetrs solution failed",intlist=(/info/),procs=procs)
      stop
   endif

! Multiply by A21 and subtract from b1

   do i=1,loc_neq
      row = eq + i-1
      do j=linear_system%begin_row(row),linear_system%end_row(row)
         col = linear_system%column_index(j)
         if (col == NO_ENTRY) cycle
         if (col >= eq .and. col < eq+loc_neq) cycle
         if (linear_system%equation_type(col) == DIRICHLET) cycle
         linear_system%rhs(col) = linear_system%rhs(col) - &
                                  linear_system%matrix_val(j)*tempvec(i,1)
      end do
   end do

! next block of equations

   eq = eq + loc_neq
end do

deallocate(tempvec)

end subroutine static_condensation_rhs

!          ---------------------
subroutine destroy_linear_system(linear_system)
!          ---------------------

!----------------------------------------------------
! This routine frees the memory used by the linear system
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type),intent(inout) :: linear_system

!----------------------------------------------------
! local variables

integer :: i, allocstat

!----------------------------------------------------
! Begin executable code

if (linear_system%coarse_band_exists) then
   call destroy_lapack_band(linear_system%coarse_matrix)
   linear_system%coarse_band_exists = .false.
endif
if (linear_system%lapack_symm_band_exists) then
   call destroy_lapack_band(linear_system%lapack_mat)
   linear_system%lapack_symm_band_exists = .false.
endif
if (linear_system%lapack_gen_band_exists) then
   call destroy_lapack_band(linear_system%lapack_mat)
   linear_system%lapack_gen_band_exists = .false.
endif
if (linear_system%petsc_matrix_exists) then
   call destroy_petsc_linear_system(linear_system%petsc_matrix)
   linear_system%petsc_matrix_exists = .false.
endif

if (associated(linear_system%begin_row)) then
   if (associated(linear_system%condensed,linear_system%stiffness)) then
      nullify(linear_system%condensed)
   endif
   if (associated(linear_system%rhs_cond,linear_system%rhs_nocond)) then
      nullify(linear_system%rhs_cond)
   endif
   if (associated(linear_system%elem_block)) then
      do i=lbound(linear_system%elem_block,dim=1), &
           ubound(linear_system%elem_block,dim=1)
         if (associated(linear_system%elem_block(i)%matrix)) then
            deallocate(linear_system%elem_block(i)%matrix, &
                       linear_system%elem_block(i)%ipiv,stat=allocstat)
            if (allocstat /= 0) then
               call warning("deallocation failed in destroy_linear_system (elem_block component)")
            endif
         endif
      end do
      deallocate(linear_system%elem_block,stat=allocstat)
      if (allocstat /= 0) then
         call warning("deallocation failed in destroy_linear_system (elem_block)")
      endif
   endif
   deallocate(linear_system%stiffness, &
              linear_system%column_index, &
              linear_system%begin_row, &
              linear_system%end_row_bubble, &
              linear_system%end_row_linear, &
              linear_system%end_row_edge, &
              linear_system%begin_level, &
              linear_system%edge_begin_degree, &
              linear_system%face_begin_degree, &
              linear_system%rhs_nocond, &
              linear_system%r_mine, &
              linear_system%r_others, &
              linear_system%need_r_others, &
              linear_system%iown, &
              linear_system%solution, &
              linear_system%equation_type, &
              linear_system%gid, &
              linear_system%hold_dirich, &
              linear_system%s_int, &
              linear_system%s_bnd, &
              linear_system%s_int_inv, &
              linear_system%s_bnd_inv, &
              stat=allocstat)
   if (allocstat /= 0) then
      call warning("deallocation failed in destroy_linear_system")
   endif
   if (associated(linear_system%end_row_face)) then
      deallocate(linear_system%end_row_face)
   endif
   nullify(linear_system%end_row)
   if (associated(linear_system%evecs)) then
      deallocate(linear_system%evecs,stat=allocstat)
      if (allocstat /= 0) then
         call warning("deallocation failed in destroy_linear_system")
      endif
   endif
   nullify(linear_system%matrix_val)
   if (associated(linear_system%mass)) then
      deallocate(linear_system%mass, stat=allocstat)
      if (allocstat /= 0) then
         call warning("deallocation failed in destroy_linear_system (mass)")
      endif
   endif
   if (associated(linear_system%condensed)) then
      deallocate(linear_system%condensed,linear_system%rhs_cond,stat=allocstat)
      if (allocstat /= 0) then
         call warning("deallocation failed in destroy_linear_system (condensed)")
      endif
   endif
   call hash_table_destroy(linear_system%eq_hash)
endif

! Destroy the linked list of vertices for floating regions

if (associated(linear_system%floating_head)) then
   deallocate(linear_system%floating_head,linear_system%floating_next)
endif

linear_system%linsys_exists = .false.

end subroutine destroy_linear_system

end module make_linsys
